import os
import time

import dash_bootstrap_components as dbc
import dash_interactive_graphviz as dg
import pandas as pd
from dash import html, dcc, Output, Input, State, dash, Dash
import dash_cytoscape as cyto
from matplotlib import pyplot as plt
from sklearn import tree

from ...conf import parameters as cfg
from . import utils
from .MetaTab import MetaTab
from ...domain import Controller
from ...service import Plots, Utils, init_logger, log_exceptions

PATH_TO_BIGRESULTS = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", "..", "..", "big_results.p")
)

CONFIG = {
    "toImageButtonOptions": {
        "format": "svg",  # one of png, svg, jpeg, webp
        "height": None,
        "width": None,
        "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
    }
}


class ResultsTab(MetaTab):
    def __init__(self, app: Dash, metabo_controller: Controller):
        super().__init__(app, metabo_controller)
        self._logger = init_logger()
        # self.r = pkl.load(open(PATH_TO_BIGRESULTS, "rb"))
        self.r = self.metabo_controller.get_all_results()
        self._plots = Plots("blues")

    def getLayout(self) -> dbc.Tab:
        __resultsMenuDropdowns = dbc.Card(
            className="results_menu_dropdowns",
            children=[
                dbc.CardBody(
                    [
                        html.Div(
                            className="dropdowns",
                            children=[
                                html.H6("Classification Design : "),
                                dbc.Select(
                                    id="design_dropdown",
                                    className="form_select",
                                    options=[{"label": "None", "value": "None"}],
                                    value="None",
                                ),
                            ],
                        ),
                        html.Div(
                            className="dropdowns",
                            children=[
                                html.H6("ML Algorithm : "),
                                dbc.Select(
                                    id="ml_dropdown",
                                    className="form_select",
                                    options=[{"label": "None", "value": "None"}],
                                    value="None",
                                ),
                            ],
                        ),
                        dbc.Button(
                            "Load",
                            color="primary",
                            id="load_ML_results_button",
                            className="custom_buttons",
                            n_clicks=0,
                        ),
                        html.Div(id="output_button_load_ML_results"),
                    ],
                    id="menu_results",
                )
            ],
        )

        __currentExperimentInfo = dbc.Card(
            children=[
                dbc.CardBody(
                    children=[
                        html.H6(
                            "Current experiment info"
                        ),  # , style={"marginTop": 25},
                        # html.Div(id="view_info", children=[
                        dcc.Loading(
                            id="loading_expe_table",
                            children=html.Div(id="expe_table", children=""),
                            type="circle",
                        ),
                        # dcc.Loading(id="loading-1", children=[html.Div(id="loading-output-1")],
                        #             type="dot", color="#13BD00")
                    ]
                )
            ],
            className="w-25",
        )

        _resultsInfo = html.Div(
            className="Results_info",
            children=[
                __resultsMenuDropdowns,
                __currentExperimentInfo,
            ],
        )

        ___pcaPlot = html.Div(
            className="pca_plot_and_title",
            children=[
                html.Div(
                    className="title_and_help",
                    children=[
                        html.H6("PCA", id="PCA_title"),
                        dcc.RadioItems(
                            id='pca_dimensions',
                            options=[
                                {'label': ' 2D', 'value': '2d'},
                                {'label': ' 3D', 'value': '3d'},
                            ],
                            value='2d',
                            labelStyle={'margin': 'auto', 'padding': '5px'},
                            inline=True
                        ),
                    ],
                    style={"display": "flex", "justify-content": "space-between"},
                ),
                # Should we put the title on the plot?
                dcc.Loading(
                    dcc.Graph(id="PCA", config=CONFIG), type="dot", color="#13BD00"
                ),
                dcc.Slider(
                    step=None,
                    value=1,
                    marks={**cfg.default_marks, **cfg.all_mark},
                    id="pca_slider",
                ),
            ],
        )

        ___umap = html.Div(
            className="umap_plot_and_title",
            children=[
                html.Div(
                    className="title_and_help",
                    children=[
                        html.H6("Umap"),
                        dcc.RadioItems(
                            id='umap_dimensions',
                            options=[
                                {'label': ' 2D', 'value': '2d'},
                                {'label': ' 3D', 'value': '3d'},
                            ],
                            value='2d',
                            # add space between radio and label
                            labelStyle={'margin': 'auto', 'padding': '5px'},
                            inline=True

                        ),
                    ],
                    style={"display": "flex", "justify-content": "space-between"},
                ),
                dcc.Loading(
                    dcc.Graph(id="umap_overview", config=CONFIG),
                    type="dot",
                    color="#13BD00",
                ),
                dcc.Slider(
                    step=None,
                    value=1,
                    marks={**cfg.default_marks, **cfg.all_mark},
                    id="umap_slider",
                )
            ],
        )

        ___2dPlot = html.Div(
            className="umap_plot_and_title",
            children=[
                html.Div(
                    className="title_and_help",
                    children=[
                        html.H6("2D"),
                        dbc.Button(
                            "[?]",
                            className="text-muted btn-secondary popover_btn",
                            id="help_2dPlot",
                        ),
                        dbc.Popover(
                            children=[dbc.PopoverBody("Blablabla wout wout")],
                            id="pop_help_2dPlot",
                            is_open=False,
                            target="help_2dPlot",
                        ),
                    ],
                ),
                dcc.Loading(
                    dcc.Graph(id="2d_overview", config=CONFIG),
                    type="dot",
                    color="#13BD00",
                ),
            ],
        )

        ___3dPlot = html.Div(
            className="umap_plot_and_title",
            children=[
                html.Div(
                    className="title_and_help",
                    children=[
                        html.H6("3D"),
                        dbc.Button(
                            "[?]",
                            className="text-muted btn-secondary popover_btn",
                            id="help_3dPlot",
                        ),
                        dbc.Popover(
                            children=[dbc.PopoverBody("Blablabla wout wout")],
                            id="pop_help_3dPlot",
                            is_open=False,
                            target="help_3dPlot",
                        ),
                    ],
                ),
                dcc.Loading(
                    dcc.Graph(id="3d_overview", config=CONFIG),
                    type="dot",
                    color="#13BD00",
                ),
            ],
        )

        __dataResultTab = dbc.Tab(
            className="sub_tab",
            label="Data",
            children=[
                html.Div(className="fig_group", children=[___pcaPlot, ___umap]),
                html.Div(className="fig_group", children=[___2dPlot, ___3dPlot]),
            ],
        )

        ___accuracyPlot = html.Div(
            className="acc_plot_and_title",
            children=[
                html.Div(
                    className="title_and_help",
                    children=[
                        html.H6("Balanced accuracy plot"),
                        dbc.Button(
                            "[?]",
                            className="text-muted btn-secondary popover_btn",
                            id="help_accPlot",
                        ),
                        dbc.Popover(
                            children=[
                                dbc.PopoverBody(
                                    "Accuracies for each split on train and test set. Here you would want to check"
                                    "the difference between each set, because a really good train performance and a mediocre"
                                    "or bad test performance is a sign of over-fitting."
                                )
                            ],
                            id="pop_help_accPlot",
                            is_open=False,
                            target="help_accPlot",
                        ),
                    ],
                ),
                dcc.Loading(
                    dcc.Graph(id="accuracy_overview", config=CONFIG),
                    type="dot",
                    color="#13BD00",
                ),
            ],
        )

        ___globalMetric = html.Div(
            className="w-25",
            children=[
                html.H6("Global confusion matrix"),
                dcc.Loading(
                    dcc.Graph(id="conf_matrix", config=CONFIG),
                    type="dot",
                    color="#13BD00",
                ),
            ],
        )
        ___specificFilters = html.Div(
            className="fig_group_col",
            children=[
                html.Div(
                    className="",
                    children=[
                        html.H6("Splits number"),
                        dbc.Select(
                            id="splits_dropdown",
                            className="form_select_large",
                            options=[{"label": "None", "value": "None"}],
                            value="None",
                        ),
                        dbc.Button(
                            "Update",
                            color="primary",
                            id="update_specific_results_button",
                            className="custom_buttons",
                            n_clicks=0,
                        ),
                        html.Div(id="output_button_update_specific_results"),
                    ],
                ),
                html.Div(
                    className="",
                    style={"display": "flex", "width": "100%"},
                    children=[
                        html.Div(
                            children=[
                                html.H6("Confusion matrix"),
                                dcc.Loading(
                                    dcc.Graph(id="split_conf_matrix", config=CONFIG),
                                    type="dot",
                                    color="#13BD00",
                                ),
                            ],
                        ),
                        html.Div(
                            children=[
                                html.H6("Table of used hyperparameter"),
                                dcc.Loading(
                                    html.Div(id="hyperparam_table", children="", style={"margin-top": "2em"}),
                                ),
                            ],
                        )
                    ],
                ),
            ],
        )
        ___metricsTable = html.Div(
            className="table_features",
            children=[
                html.H6("Metrics table : mean(std)"),
                dcc.Loading(
                    id="loading_metrics_table",
                    children=html.Div(id="metrics_score_table", children=""),
                    type="circle",
                ),
            ],
        )

        __algoResultsTab = dbc.Tab(
            className="sub_tab",
            label="Algorithm",
            children=[
                html.Div(
                    className="fig_group",
                    children=[
                        ___accuracyPlot,
                        # ___globalMetric,
                        ___metricsTable,
                    ],
                ),
                html.Div(className="fig_group", children=[___specificFilters]),
            ],
        )

        __DTTreeTab = dbc.Tab(
            id="DTTT", className="sub_tab", label="DT Tree", disabled=True,
            children=[
                html.Div(dg.DashInteractiveGraphviz(
                    id="DTTT_graph"
                ),
                    style={"letter-spacing": "0"}),
            ]
        )

        ___featuresTable = html.Div(
            className="table_features",
            children=[
                html.H6("Top 10 features sorted by importance"),
                dbc.Button(
                    "Export",
                    color="primary",
                    id="export_features",
                    className="custom_buttons",
                    n_clicks=0,
                ),
                dcc.Download(id="download_dataframe_csv"),
                dcc.Loading(
                    id="loading_features_table",
                    children=html.Div(id="features_table", children=""),
                    type="circle",
                ),
            ],
        )
        ___stripChart = html.Div(
            className="umap_plot_and_title",
            children=[
                html.Div(
                    className="title_and_help",
                    children=[
                        html.H6("StripChart of features"),
                        dbc.Button(
                            "[?]",
                            className="text-muted btn-secondary popover_btn",
                            id="help_stripChart",
                        ),
                        dbc.Popover(
                            children=[dbc.PopoverBody("Blablabla wout wout")],
                            id="pop_help_stripChart",
                            is_open=False,
                            target="help_stripChart",
                        ),
                    ],
                ),
                # dbc.Select(
                #     id="features_dropdown",
                #     className="form_select",
                #     options=[{"label": "None", "value": "None"}],
                #     value="None",
                #     style={"width": "35%"},
                # ),
                dcc.Loading(
                    dcc.Graph(id="features_stripChart", config=CONFIG),
                    type="dot",
                    color="#13BD00",
                ),
                dcc.Slider(
                    step=None,
                    value=1,
                    marks=cfg.default_marks,
                    id="strip_chart_slider",
                )
            ],
        )

        ___coocMatrix = html.Div(
            id="cooc_matrix",
            style={"width": "50%", "height": "5em"},
            children=[
                html.H6("Co-occurence graph of features"),
                dcc.Loading(
                    children=[
                        cyto.Cytoscape(
                            id="cooc_matrix_graph",
                            layout={"name": "cose"},
                            style={"width": "100%", "height": "100%", "border": "1px solid black"},
                            stylesheet=self._plots.get_default_stylesheet_for_cooc_graph(),
                            elements=[],

                        ),
                        html.Div(
                            id="cooc_matrix_graph_error",
                            children="",
                            style={"color": "red", "height": "100%", "background-color": "lightgrey",
                                   "text-align": "center", "vertical-align": "middle"},
                        ),
                    ],
                    type="dot",
                    color="#13BD00",
                ),
            ],
        )

        __featuresResultsTab = dbc.Tab(
            className="sub_tab",
            label="Features",
            children=[
                html.Div(
                    className="fig_group", children=[___featuresTable,
                                                     ___stripChart
                                                     ]
                ),
                ___coocMatrix
            ],
        )

        _mainPlotContent = html.Div(
            id="main_plots-content",
            children=[  # className="six columns",
                dbc.Tabs(
                    className="custom_sub_tabs",
                    id="sub_tabs",
                    children=[
                        __dataResultTab,
                        __algoResultsTab,
                        __featuresResultsTab,
                        __DTTreeTab,
                    ],
                )
            ],
        )

        return dbc.Tab(
            className="global_tab",
            id="results_tab",
            label="Results",
            children=[
                dcc.Store(id='design_dropdown_store', storage_type='session'),
                dcc.Store(id='ml_dropdown_store', storage_type='session'),
                dcc.Store(id='splits_dropdown_store', storage_type='session'),
                _resultsInfo,
                _mainPlotContent
            ],
        )

    def _registerCallbacks(self) -> None:
        @self.app.callback(
            Output("pop_help_accPlot", "is_open"),
            [Input("help_accPlot", "n_clicks")],
            [State("pop_help_accPlot", "is_open")],
        )
        def toggle_popover(n, is_open):
            if n:
                return not is_open
            return is_open

        @self.app.callback(
            [Output("design_dropdown", "options"), Output("design_dropdown", "value")],
            [Input("custom_big_tabs", "active_tab")],
            State('design_dropdown_store', 'data'),
        )
        @log_exceptions(self._logger)
        def update_results_dropdown_design(active, stored_value):
            if active == "tab-3":
                self.r = self.metabo_controller.get_all_results()
                experiment_designs = list(self.r.keys())
                if len(experiment_designs) == 0:
                    return dash.no_update, dash.no_update
                if stored_value is not None and stored_value in experiment_designs:
                    return Utils.format_list_for_checklist(experiment_designs), stored_value
                return (
                    Utils.format_list_for_checklist(experiment_designs),
                    experiment_designs[0],
                )
            else:
                return dash.no_update, dash.no_update

        @self.app.callback(
            [Output("ml_dropdown", "options"), Output("ml_dropdown", "value")],
            [Input("design_dropdown", "value")],
            [State("custom_big_tabs", "active_tab"), State('ml_dropdown_store', 'data')],
        )
        @log_exceptions(self._logger)
        def update_results_dropdown_algo(design, active, stored_value):
            if active == "tab-3":
                a = list(self.r[design].keys())
                if stored_value is not None and stored_value in a:
                    return [{"label": i, "value": i} for i in a], stored_value
                else:
                    return [{"label": i, "value": i} for i in a], a[0]
            else:
                return dash.no_update

        @self.app.callback(
            [Output("splits_dropdown", "options"), Output("splits_dropdown", "value")],
            [Input("sub_tabs", "active_tab")],
            [State("ml_dropdown", "value"), State("design_dropdown", "value"), State('splits_dropdown_store', 'data')],
        )
        @log_exceptions(self._logger)
        def update_nbr_splits_dropdown(active, algo, design, stored_value):
            if active == "tab-1":
                a = list(self.r[design][algo].splits_number)
                if stored_value is not None and stored_value in a:
                    return [{"label": i, "value": i} for i in a], stored_value
                else:
                    return [{"label": i, "value": i} for i in a], a[0]
            else:
                return dash.no_update

        @self.app.callback(
            [Output("loading-output-1", "children")],
            [Input("custom_big_tabs", "active_tab")],
        )
        def input_triggers_spinner(value):
            time.sleep(1)
            return

        # --- Callbacks to Save Values (persist dropdown selection) ---
        @self.app.callback(
            Output('design_dropdown_store', 'data'),
            Input('load_ML_results_button', 'n_clicks'),
            State('design_dropdown', 'value'),
        )
        def save_design_dropdown_value(_, value):
            return value

        @self.app.callback(
            Output('ml_dropdown_store', 'data'),
            Input('load_ML_results_button', 'n_clicks'),
            State('ml_dropdown', 'value'),
        )
        def save_ml_dropdown_value(_, value):
            return value

        @self.app.callback(
            Output('splits_dropdown_store', 'data'),
            Input('update_specific_results_button', 'n_clicks'),
            State('splits_dropdown', 'value'),
        )
        def save_splits_dropdown_value(_, value):
            return value

        # @self.app.callback(
        #     Output("2features", "figure"),
        #     [Input("load_ML_results_button", "n_clicks"), Input("pca_slider", "value")],
        #     [State("ml_dropdown", "value"), State("design_dropdown", "value")],
        # )
        # def show_pca(n_clicks, features, algo, design_name):
        #     if n_clicks >= 1:
        #         df = self.r[design_name][algo].results["features_table"]
        #         classes = self.r[design_name][algo].results["classes"]
        #         return self._plots.show_two_most_important_feature(df[pca_value], classes, pca_value, algo)
        #     else:
        #         return dash.no_update

        @self.app.callback(
            [
                Output("cooc_matrix_graph", "elements"),
                Output("cooc_matrix_graph", "style"),
                Output("cooc_matrix_graph_error", "children")
            ],
            [
                Input("load_ML_results_button", "n_clicks")
            ],
            [
                State("ml_dropdown", "value"), 
                State("design_dropdown", "value")
            ]
        )
        @log_exceptions(self._logger)
        def show_cooc_matrix(_, algo, design_name):
            if algo == "None" or design_name == "None":
                return [dash.no_update] * 3
            
            (
                counter, 
                mean_importance, 
                number_of_split, 
                cardinality
            ) = self.r[design_name][algo].results["coocurence_matrix"]

            if counter is None:
                msg = "Due to the high cardinality of the features, " \
                "the co-occurrence graph cannot be displayed. " \
                "The estimated cardinality is " + str(int(cardinality)) + \
                " and exceed the 1000 links limit."
                return dash.no_update, {'display': 'none'}, msg
                   
            parameters = self._plots.create_coocurence_graph(counter, mean_importance, number_of_split)
            return parameters, {'display': 'block', "width": "100%", "height": "800px"}, ""

        @self.app.callback(Output('cooc_matrix_graph', 'stylesheet'),
                           [Input('cooc_matrix_graph', 'selectedNodeData')])
        @log_exceptions(self._logger)
        def update_stylesheet(nodes):
            default_stylesheet = self._plots.get_default_stylesheet_for_cooc_graph()
            if nodes is None:
                return default_stylesheet
            else:
                updated_stylesheet = default_stylesheet.copy()
                self._logger.info(f"nodes:\n{nodes}")
                for node in nodes:
                    updated_stylesheet.append(self._plots.format_style_for_selected_node(node))
                return updated_stylesheet

        @self.app.callback(
            [
                Output("pca_slider", "marks"),
                Output("umap_slider", "marks"),
                Output("strip_chart_slider", "marks"),
                Output("pca_slider", "value"),
                Output("umap_slider", "value"),
                Output("strip_chart_slider", "value"),
            ],
            [
                Input("load_ML_results_button", "n_clicks")
            ],
            [
                State("ml_dropdown", "value"), State("design_dropdown", "value")
            ]
        )
        @log_exceptions(self._logger)
        def update_sliders_with_used(_, algo, design_name):
            """Update PCA, UMAP and Strip chart sliders marks and triggers each plots by updating there values"""
            if algo == "None" or design_name == "None":
                return [dash.no_update] * 6
            
            feature_df = self.r[design_name][algo].results["features_table"]
            number_of_used_feature = len(feature_df[feature_df["times_used"] > 0])

            marks_container = []
            locations_container = []

            marks, used_location = utils.update_marks(
                custom_value=number_of_used_feature, 
                add_all_value=True
            )
            pca_marks, umap_marks = marks, marks
            pca_location, umap_location = used_location, used_location
            marks_container.extend([pca_marks, umap_marks])
            locations_container.extend([pca_location, umap_location])

            strip_chart_marks, strip_location = utils.update_marks(
                custom_value=number_of_used_feature, 
                add_all_value=False
            )
            marks_container.append(strip_chart_marks)
            locations_container.append(strip_location)

            return (*marks_container, *locations_container)
                    

        @self.app.callback(
            [
                Output("PCA", "figure")
            ],
            [
                Input("pca_slider", "value"),
                Input("pca_dimensions", "value")
            ],
            [
                State("ml_dropdown", "value"),
                State("design_dropdown", "value"),
                State("pca_slider", "marks")
            ],
            prevent_initial_call=True
        )
        @log_exceptions(self._logger)
        def show_pca(pca_value, dimensions, algo, design_name, marks):
            """
            pca_value : represent the number of feature selected by the slider, but is given as indexes
            """
            if algo == "None" or design_name == "None":
                return dash.no_update

            classes = self.r[design_name][algo].results["classes"]

            index = utils.get_index_from_marks(pca_value, marks)

            if dimensions == "2d":
                data_list, labels_list = self.r[design_name][algo].results["pca_data"]
                fig = self._plots.show_PCA(
                    data_list[index], labels_list[index], classes, index, algo,
                    self.r[design_name][algo].results["samples_id"]
                )
            elif dimensions == "3d":
                data_list, labels_list = self.r[design_name][algo].results["3d_pca_data"]
                fig = self._plots.show_3D_PCA(
                    data_list[index], labels_list[index], classes, index, algo,
                    self.r[design_name][algo].results["samples_id"]
                )
            
            return [fig]

        @self.app.callback(
            [
                Output("umap_overview", "figure")
            ],
            [
                Input("load_ML_results_button", "n_clicks"),
                Input("umap_slider", "value"),
                Input("umap_dimensions", "value"),
            ],
            [
                State("ml_dropdown", "value"), 
                State("design_dropdown", "value"), 
                State("umap_slider", "marks")
            ],
        )
        @log_exceptions(self._logger)
        def show_umap(_, slider_value, dimensions, algo, design_name, marks):
            if algo == "None" or design_name == "None":
                return dash.no_update
            
            classes = self.r[design_name][algo].results["classes"]

            index = utils.get_index_from_marks(slider_value, marks)

            if dimensions == "2d":
                df = self.r[design_name][algo].results["umap_data"]
                fig = self._plots.show_umap(
                    df[index], classes, algo, index, 
                    self.r[design_name][algo].results["samples_id"]
                )
            elif dimensions == "3d":
                df = self.r[design_name][algo].results["3d_umap_data"]
                fig = self._plots.show_3D_umap(
                    df[index], classes, algo, index, 
                    self.r[design_name][algo].results["samples_id"]
                )
            
            return [fig]
    

        @self.app.callback(
            [
                Output("2d_overview", "figure")
            ],
            [
                Input("load_ML_results_button", "n_clicks")
            ],
            [
                State("ml_dropdown", "value"), 
                State("design_dropdown", "value")
            ],
        )
        @log_exceptions(self._logger)
        def show_2d(_, algo, design_name):
            if algo == "None" or design_name == "None":
                return dash.no_update
            
            df = (
                self.r[design_name][algo].results["features_2d_and_3d"].iloc[:, :-1]
            )
            classes = self.r[design_name][algo].results["classes"]
            fig = self._plots.show_2d(
                df, classes, 
                self.r[design_name][algo].results["samples_id"]
            )
            return [fig]
            

        @self.app.callback(
            [
                Output("3d_overview", "figure")
            ],
            [
                Input("load_ML_results_button", "n_clicks")
            ],
            [
                State("ml_dropdown", "value"), 
                State("design_dropdown", "value")
            ],
        )
        @log_exceptions(self._logger)
        def show_3d(_, algo, design_name):
            if algo == "None" or design_name == "None":
                return dash.no_update
            
            df = self.r[design_name][algo].results["features_2d_and_3d"]
            classes = self.r[design_name][algo].results["classes"]
            fig = self._plots.show_3d(
                df, classes, 
                self.r[design_name][algo].results["samples_id"]
            )
            return [fig]

        @self.app.callback(
            [
                Output("expe_table", "children")
            ],
            [
                Input("load_ML_results_button", "n_clicks")
            ],
            [
                State("ml_dropdown", "value"), 
                State("design_dropdown", "value")
            ],
        )
        @log_exceptions(self._logger)
        def get_experiment_statistics(_, algo, design_name):
            if algo == "None" or design_name == "None":
                return dash.no_update
            
            df = self.r[design_name][algo].results["info_expe"]
            table_body = self._plots.show_exp_info_all(df)
            table = dbc.Table(
                table_body, id="table_exp_info", borderless=True, hover=True
            ) # dbc.Table.from_dataframe(df, borderless=True)
            return [table]


        @self.app.callback(
            [
                Output("accuracy_overview", "figure")
            ],
            [
                Input("load_ML_results_button", "n_clicks")
            ],
            [
                State("ml_dropdown", "value"), 
                State("design_dropdown", "value")
            ],
        )
        @log_exceptions(self._logger)
        def generates_accuracyPlot_global(_, algo, design_name):
            if algo == "None" or design_name == "None":
                return dash.no_update
            
            df = self.r[design_name][algo].results["accuracies_table"]
            fig = self._plots.show_accuracy_all(df, algo)
            return [fig]

        @self.app.callback(
            [
                Output("metrics_score_table", "children")
            ],
            [
                Input("load_ML_results_button", "n_clicks")
            ],
            [
                State("ml_dropdown", "value"), 
                State("design_dropdown", "value")
            ],
        )
        @log_exceptions(self._logger)
        def show_metrics(_, algo, design_name):
            if algo == "None" or design_name == "None":
                return dash.no_update
            
            df = self.r[design_name][algo].results["metrics_table"]
            table = dbc.Table.from_dataframe(df, borderless=True)
            return [table]


        @self.app.callback(
            [
                Output("split_conf_matrix", "figure"),
                Output("hyperparam_table", "children")
            ],
            [
                Input("update_specific_results_button", "n_clicks")],
            [
                State("ml_dropdown", "value"),
                State("design_dropdown", "value"),
                State("splits_dropdown", "value"),
            ],
        )
        @log_exceptions(self._logger)
        def compute_split_conf_matrix(_, algo, design_name, split):
            if algo == "None" or design_name == "None":
                return dash.no_update, dash.no_update
            
            cm = self.r[design_name][algo].results[split]["Confusion_matrix"][1]
            labels = self.r[design_name][algo].results[split]["Confusion_matrix"][0]

            text_mat = []
            for i, line in enumerate(cm):
                text_mat.append([])
                for j, col in enumerate(line):
                    text_mat[i].append(str(col))

            hps = self.r[design_name][algo].results[split]["hyperparameters"]

            hps_df = pd.DataFrame.from_dict({"Hyperparameters": hps.keys(), "Values": hps.values()})

            dash_table_element = dbc.Table.from_dataframe(hps_df, borderless=True)
            fig = self._plots.show_general_confusion_matrix(
                cm, labels, text_mat, algo, split
            )

            return  fig, dash_table_element


        @self.app.callback(
            Output("features_table", "children"),
            Input("load_ML_results_button", "n_clicks"),
            [
                State("ml_dropdown", "value"), 
                State("design_dropdown", "value")
            ],
        )
        @log_exceptions(self._logger)
        def show_features(_, algo, design_name):
            if algo == "None" or design_name == "None":
                return dash.no_update
            
            df = self.r[design_name][algo].results["features_table"].copy()
            df = df.sort_values(by="importance_usage", ascending=False)
            df = df.round(4)

            return dbc.Table.from_dataframe(df.iloc[:10, :], borderless=True)


        @self.app.callback(
            Output("download_dataframe_csv", "data"),
            [Input("export_features", "n_clicks")],
            [State("ml_dropdown", "value"), State("design_dropdown", "value")],
            prevent_initial_call=True,
        )
        @log_exceptions(self._logger)
        def export_download_features_table(n_click, algo, design_name):
            if n_click >= 1:
                df = self.r[design_name][algo].results["features_table"]
                return dcc.send_data_frame(
                    df.to_csv, "featuresImportancesTable" + algo + ".csv"
                )
            else:
                return dash.no_update


        @self.app.callback(
            [
                Output("features_stripChart", "figure")
            ],
            [
                Input("strip_chart_slider", "value")
            ],
            [
                State("ml_dropdown", "value"), 
                State("design_dropdown", "value"), 
                State("strip_chart_slider", "marks")
            ],
        )
        @log_exceptions(self._logger)
        def show_stripChart_features(slider_value, algo, design_name, marks):
            if algo == "None" or design_name == "None":
                return dash.no_update

            try:
                real_value = utils.get_index_from_marks(slider_value, marks)
                strip_chart_data = self.r[design_name][algo].results["features_stripchart"][real_value]
                fig = self._plots.show_metabolite_levels(
                    strip_chart_data, algo, 
                    self.r[design_name][algo].results["samples_id"]
                )
                return [fig]
            except IndexError:
                return dash.no_update

        @self.app.callback(
            [
                Output("DTTT", "disabled"), 
                Output("DTTT_graph", "dot_source")
            ],
            [
                Input("load_ML_results_button", "n_clicks")
            ],
            [
                State("ml_dropdown", "value"), 
                State("design_dropdown", "value")
            ],
        )
        @log_exceptions(self._logger)
        def disable_DTTT(_, algo, design_name):
            if algo == "None" or design_name == "None":
                return dash.no_update, dash.no_update
            
            if algo == "DecisionTree":
                model = self.r[design_name][algo].results["best_model"]
                classes = list(set(self.r[design_name][algo].results["classes"]))
                plt.margins(0.05)
                df = self.r[design_name][algo].results["features_table"]
                df.sort_index(inplace=True)
                features_name = list(df["features"])
                dot_data = tree.export_graphviz(
                    model,
                    out_file=None,
                    class_names=classes,
                    feature_names=features_name,
                    proportion=True,
                    filled=True,
                    rounded=True,
                    special_characters=True,
                )

                return False, dot_data
            
            return True, ""



import numpy as np

from ...conf import parameters as cfg


def update_marks(custom_value: int, add_all_value: bool = False) -> dict:
    """
    Update the marks of the features slider to include a custom value.
    
    If the custom value is already in the default marks, it will be used and
    marked as "(used)".
    
    If the custom value is not in the default marks, it will be added to the
    marks at a position that is the closest to its value.
    
    If add_all_value is True, the "All" mark will be added to the marks.
    
    Returns a tuple containing the updated marks and the location of the custom
    value.
    """
    locations = [int(l) for l in cfg.default_marks.keys()]
    values = np.array([int(v) for v in cfg.default_marks.values()])
    
    # If the custom value is already there, we use it a used feature
    if custom_value in values:
        marks = cfg.default_marks.copy()
        idx = np.where(values == custom_value)[0].item()
        location_key = str(locations[idx])
        marks[location_key] = marks[location_key] + " (used)"
        
        if add_all_value:
            marks.update(cfg.all_mark)
        
        return marks, location_key
    
    greater_indicator = values > custom_value
    if np.all(greater_indicator):
        # The value is on the far left
        custom_location = locations[0] - cfg.custom_mark_offset
    elif not np.any(greater_indicator):
        # The value is on the far right
        custom_location = locations[-1] + cfg.custom_mark_offset
    else:
        greater_idx = np.argmax(greater_indicator)
        distance_from_greater = values[greater_idx] - custom_value
        distance_from_smaller = custom_value - values[greater_idx - 1]
        if distance_from_smaller < distance_from_greater:
            # Closer to the left, we will add the offset on the left mark
            custom_location = locations[greater_idx - 1] + cfg.custom_mark_offset
        elif distance_from_smaller == distance_from_greater:
            # Equal distance
            custom_location = (locations[greater_idx - 1] + locations[greater_idx]) / 2
        else:
            # Closer the the right, we will remove the offset from the right mark
            custom_location = locations[greater_idx] - cfg.custom_mark_offset
    
    marks = cfg.default_marks.copy()
    value_label = {
        'label': str(custom_value) + ' (used)', 
        'style': {
            "top": "-3em"
        }
    }
    marks[str(custom_location)] = value_label

    if add_all_value:
        marks = {**marks, **cfg.all_mark}

    return marks, custom_location



def get_index_from_marks(number: float, marks: dict) -> int:
    """This is really specific to how the results are saved. 
    See pico.domain.Results._produce_PCA (or _produce_UMAP).

    The first results are from cfg.features, then used_features, then all.

    The result differs when there is "All" features (e.g. strip chart doesnt not include All).
    """
    contain_all = False
    for v in marks.values():
        if isinstance(v, str) and v == "All":
            contain_all = True

    match marks[str(number)]:
        case dict(value):
            # We have the used feature case, index is -2 or -1 when 
            index = -2 if contain_all else -1
        case "All":
            # We have the "All" case, index is -1
            index = -1
        case str(value) if "used" in value:
            # We have the feature case
            index = -2 if contain_all else -1
        case _:
            # The number is the index
            index = int(number)


    return index



from .SplitGroup import SplitGroup
from .MetaData import MetaData
from .ClassificationDesign import ClassificationDesign
from .MetaboExperiment import MetaboExperiment
from .ModelFactory import ModelFactory
from .Results import Results
from .Controller import Controller
from .DataMatrix import DataMatrix
from .MetaboModel import MetaboModel



import subprocess
import sys
import time
import webbrowser
import threading

from typing import Annotated, Optional

import typer

from . import console
from .. import __version__
from .. import app as pico_app

# Use print from `rich` library
print = console.print

app = typer.Typer()

PYPI_PROJECT_NAME = "pico-ml"

@app.callback(invoke_without_command=True, no_args_is_help=True)
def main(
    version: bool = typer.Option(False, "--version", "-v", help="PICO CLI version")
):
    """
    PICO CLI!
    """
    if version:
        typer.echo(__version__)
        typer.Exit()


@app.command(name="ui")
def ui(
    port: Annotated[Optional[int], typer.Option("--port", "-p", help="Specify the port of the app")] = 5000,
    verbose: Annotated[Optional[str], typer.Option("--verbose", "-v", help="Specifiy verbosity")] = "debug"
):
    """Launch the web application"""
    print("verbose", verbose)
    page_url = f"http://127.0.0.1:{port}"

    def startWebPage(url, wait_time: int = 0):
        """Open a browser on localhost at a url after a given wait_time"""
        time.sleep(wait_time)
        webbrowser.open(url)
    
    start_web_page_thread = threading.Thread(
        target=startWebPage,
        args=(page_url, 5)
    )
    start_web_page_thread.start()
    pico_app.run(debug=False, host='0.0.0.0', port=port)


@app.command()
def update():
    """Update the pico project to the latest version"""

    # See https://stackoverflow.com/a/50255019 for installation of a python package
    pip_infos = (
        subprocess.run(
            [sys.executable, "-m", "pip", "show", PYPI_PROJECT_NAME],
            check=True,
            capture_output=True,
        )
        .stdout.decode('latin-1')
        .split("\n")
    )

    editable_identifier_str = "Editable project location: "
    editable_project_location = [l for l in pip_infos if editable_identifier_str in l]
    if editable_project_location:
        print(
            "[cyan]pico[/cyan] is installed in editable mode, it should already be up to date."
        )
        print(f"Current version: {__version__}")
        return

    # pico was installed with pipx
    if "pipx" in sys.executable:
        print("[cyan]pico[/cyan] is managed by [cyan]pipx[/cyan]")
        update_function = [
            "pipx",
            "upgrade",
            PYPI_PROJECT_NAME
        ]
        print("The following command will update [cyan]pico[/cyan]:")
        print(" ".join(update_function), end="\n\n")
        typer.confirm("Do you want to proceed?", abort=True)

        # pipx is verbose enough
        subprocess.run(update_function, check=True)
        return

    # Update pico with the current Python interpreter
    update_function = [
        sys.executable,
        "-m",
        "pip",
        "install",
        "-U",
        PYPI_PROJECT_NAME
    ]
    print("The following command will update [cyan]pico[/cyan]:")
    print(" ".join(update_function), end="\n\n")
    typer.confirm("Do you want to proceed?", abort=True)

    subprocess.run(update_function, check=True, capture_output=True)
    new_version = subprocess.run(
        [sys.executable, "-m", "pico", "--version"], capture_output=True
    ).stdout.decode().strip()
    print(f"[cyan]pico[/cyan] updated to {new_version}")


# This is expose for documentation purposes
typer_click_object = typer.main.get_command(app)



import dash
import dash_bootstrap_components as dbc
from dash import html, dcc, Output, Input, State, callback_context, Dash

from .MetaTab import MetaTab
from ...service import decode_pickle_from_base64, Utils, init_logger, log_exceptions
from ...domain import Controller


class InfoTab(MetaTab):
    def __init__(self, app: Dash, metabo_controller: Controller):
        super().__init__(app, metabo_controller)
        self._logger = init_logger()
        
    def getLayout(self) -> dbc.Tab:
        _docLink = dbc.Card(
            className="cards_info",
            children=[
                dbc.CardHeader("Documentation link"),
                dbc.CardBody(
                    [
                        "You can find the official documentation ",
                        html.A(
                            href="https://elinaff.github.io/PICO/",
                            target="_blank",
                            rel="noreferrer noopener",
                            children="here",
                        ),
                        ".",
                    ]
                ),
            ],
        )

        _splitsInfo = dbc.Card(
            className="cards_info",
            children=[
                dbc.CardHeader("Splits"),
                dbc.CardBody(
                    [
                        html.P(
                            "The Splits tab is where settings to run a machine learning experiment are decided and stored into a file. "
                            "At this step, you can decide the classification designs that you wish to explore, the number of splits, the sample pairing, etc. "
                            "In the case where you have 3 or more different classes (eg: healthy, sick, extremely sick or diet1, diet2, diet3, diet4), "
                            "you will be able to combine them in two groups (designated by labels) to compare the different conditions in different ways. "
                            "For example, can the algorithms discriminate between (diet1) vs (diet2, diet3 and diet4), or (diet1 and diet2) vs (diet3 and diet4) ?"
                        ),
                        html.P(
                            "There is a hash mecanism in place to ensure that the locally saved data fits an experiment file "
                            "that might be loaded into the PICO in the future. It is computed and stored in the file of settings. This mecanism can be compared to a lock and key mecanism. The key "
                            "to check a file will only fit this particular file. It ensures reproductibility reliability."
                        ),
                    ]
                ),
            ],
        )

        _MLInfo = dbc.Card(
            className="cards_info",
            children=[
                dbc.CardHeader("Machine Learning"),
                dbc.CardBody(
                    [
                        html.P(
                            "In this tab, you select the cross-validation algorithm and its parameters and the machine learning algorithms you wish to apply on your data."
                            "The cross-validation step is usefull to optimize the algorithm : it runs several time the algorithm on small"
                            "part of the dataset with different parameters, and will keep the best model which will be applied "
                            "to your real dataset and gives you the final results and analyses."
                        ),
                        html.P(
                            "For the algorithms selection, there is some already implemented by default in the tool, so you can simply "
                            "select them. But there is also the possibility to import manually other algorithms from Scikit-Learn. In this "
                            "case you need to provide several information about the algorithm you wish to add so it can be integrated in the "
                            "analysis."
                        ),
                        html.P(
                            "The possibility to add a completely custom algorithm is also available. But it requires "
                            "modifications directly in the code files. It is meant for people with more programming abilities."
                        ),
                    ]
                ),
            ],
        )

        _resultInfo = dbc.Card(
            className="cards_info",
            children=[
                dbc.CardHeader("Results"),
                dbc.CardBody(
                    [
                        html.P(
                            "The entire section of Results is based on the perspective of analysing the features selected "
                            "by the algorithms. For one classification design, multiple algorithms can be run. "
                            "Then, the results and performances of each algorithm can be explored one by one to ensure that "
                            "the prediction and therefore the selection of the features is valid. At the end, there is a "
                            "section that aggregates the results of all the algorithms to check for redundancies. The repeated use of an "
                            "ion(feature) across different algorithm is a good indicator of the relevance of this molecule.",
                            className="card-text",
                        ),
                    ]
                ),
            ],
        )

        _loadExpe = dbc.Card(
            className="cards_info",
            style={"margin-left": "2em"},
            children=[
                dbc.CardHeader("Load MetaboExperiment"),
                dbc.CardBody(
                    [
                        html.P(
                            "You can load the file of a previous experiment and resume your analysis."
                        ),
                        dcc.Upload(
                            id="load_expe",
                            style={"width": "fit-content"},
                            children=[
                                dbc.Button(
                                    "Select File",
                                    id="load_expe_button",
                                    className="custom_buttons",
                                    color="primary",
                                )
                            ],
                        ),
                    ]
                ),
            ],
        )

        _infoFigure = dbc.Card(
            className="card_body_fig",
            children=[
                # dbc.Card("Amazing figure here", className="card_body_fig", body=True),
                dbc.CardImg(src="/assets/update_figure_steps_MeDIC_4.svg", bottom=True)
            ],
        )
        # TODO : add the filename
        _modal = dbc.Modal(
            children=[
                dbc.ModalHeader(
                    style={"padding": "2rem 3rem"},
                    children=[html.H6("Warning: Saved local files do not match")],
                ),
                dbc.ModalBody(
                    style={"padding": "2em"},
                    children=[
                        html.P(
                            "The data and/or metadata used in the MetaboExperiment file are not the same as the "
                            "local ones."
                        ),
                        html.P(
                            "Please restore the correct data and/or metadata if you want "
                            "to continue the same experiment. (Full restore)"
                        ),
                        html.P(
                            "Otherwise, you can use the same parameters with new data and/or metadata (Partial restore) "
                            "but the sample column name, target column name and classification designs won't be "
                            "restored."
                        ),
                        html.P(
                            "If you only want to see the results, it will be available (Load results) but metadata and data "
                            "matrix will be reset, as well as the classification designs."
                        ),
                        html.Div(
                            children=[
                                dcc.Upload(
                                    id="upload_datatable_modal",
                                    children=[
                                        dbc.Button(
                                            "Upload Data Matrix",
                                            id="upload_datatable_modal_button",
                                            # className="custom_buttons",
                                            color="outline-primary",
                                        )
                                    ],
                                ),
                                dcc.Loading(
                                    id="upload_datatable_modal_loading",
                                    type="dot",
                                    color="#13BD00",
                                    children=[
                                        html.Div(id="upload_datatable_modal_output"),
                                    ],
                                ),
                            ],
                            style={"display": "flex", "align-items": "center"},
                        ),
                        html.Div(
                            children=[
                                dcc.Upload(
                                    id="upload_metadata_modal",
                                    children=[
                                        dbc.Button(
                                            "Upload Metadata",
                                            id="upload_metadata_modal_button",
                                            # className="custom_buttons",
                                            color="outline-primary",
                                        )
                                    ],
                                ),
                                dcc.Loading(
                                    id="upload_metadata_modal_loading",
                                    type="dot",
                                    color="#13BD00",
                                    children=[
                                        html.Div(id="upload_metadata_modal_output"),
                                    ],
                                ),
                            ],
                            style={"display": "flex", "align-items": "center"},
                        ),
                        html.Div(
                            id="upload_datatable_modal_error_output",
                            style={"color": "red"},
                        ),
                    ],
                ),
                dbc.ModalFooter(
                    style={"padding-left": "1em"},
                    children=[
                        dbc.Button(
                            "Close", id="close", className="custom_buttons", n_clicks=0
                        ),
                        # NO FILES
                        dbc.Button(
                            "Load results",
                            id="loadAnyway",
                            className="custom_buttons push",
                            n_clicks=0,
                        ),
                        # require files
                        dbc.Button(
                            "Partial restore",
                            id="partialRestore",
                            className="custom_buttons",
                            n_clicks=0,
                        ),
                        # require files
                        dbc.Button(
                            "Full restore",
                            id="fullRestore",
                            className="custom_buttons",
                            n_clicks=0,
                        ),
                    ],
                ),
            ],
            id="warning-not-match",
            size="lg",
            is_open=False,
        )

        _hidden_div = html.Div(id="hidden_div", style={"display": "none"})

        return dbc.Tab(
            className="global_tab",
            tab_style={"margin-left": "auto"},
            label="Home",
            children=[
                _modal,
                html.Div(
                    className="fig_group",
                    children=[
                        html.Div(
                            className="column_content",
                            # WARNING !! : _infoFigure is not with the card, it's in a separate column
                            children=[
                                _docLink,
                                _splitsInfo,
                                _MLInfo,
                                _resultInfo,
                                _hidden_div,
                            ],
                        ),
                        html.Div(
                            className="column_content",
                            children=[
                                _loadExpe,
                                _infoFigure,
                            ],
                        ),
                    ],
                ),
            ],
        )  # _interpretInfo

    def _registerCallbacks(self) -> None:
        @self.app.callback(
            [
                Output("upload_datatable_modal_output", "children"),
                Output("upload_datatable_modal_output", "style"),
            ],
            [Input("upload_datatable_modal", "contents")],
            [State("load_expe", "contents")],
        )
        @log_exceptions(self._logger)
        def set_data_matrix_in_modal(contents, dto_contents):
            if contents is not None:
                metabo_experiment_dto = decode_pickle_from_base64(dto_contents)
                if Utils.is_data_the_same(contents, metabo_experiment_dto):
                    return "Data matrix uploaded successfully", {"color": "green"}
                else:
                    return (
                        "Data matrix uploaded but not corresponding to the local one",
                        {"color": "yellow"},
                    )
            else:
                return dash.no_update, dash.no_update

        @self.app.callback(
            [
                Output("upload_metadata_modal_output", "children"),
                Output("upload_metadata_modal_output", "style"),
            ],
            [Input("upload_metadata_modal", "contents")],
            [State("load_expe", "contents")],
        )
        @log_exceptions(self._logger)
        def set_metadata_in_modal(contents, dto_contents):
            if contents is not None:
                metabo_experiment_dto = decode_pickle_from_base64(dto_contents)
                if Utils.is_metadata_the_same(contents, metabo_experiment_dto):
                    return "Metadata uploaded successfully", {"color": "green"}
                else:
                    return "Metadata uploaded but not corresponding to the local one", {
                        "color": "yellow"
                    }
            else:
                return dash.no_update, dash.no_update

        @self.app.callback(
            [
                Output("warning-not-match", "is_open"),
                Output("hidden_div", "children"),
                Output("upload_datatable_modal_error_output", "children"),
            ],
            [
                Input("close", "n_clicks"),
                Input("loadAnyway", "n_clicks"),
                Input("partialRestore", "n_clicks"),
                Input("fullRestore", "n_clicks"),
                Input("load_expe", "filename"),
            ],
            [
                State("load_expe", "contents"),
                State("upload_datatable_modal", "contents"),
                State("upload_metadata_modal", "contents"),
                State("upload_datatable_modal", "filename"),
                State("upload_metadata_modal", "filename"),
            ],
        )
        @log_exceptions(self._logger)
        def toggle_modal(
            close,
            load_anyway,
            partial_restore,
            full_restore,
            filename_loaded,
            contents_loaded,
            new_data,
            new_metadata,
            new_data_name,
            new_metadata_name,
        ):
            triggered_id = callback_context.triggered[0]["prop_id"].split(".")[0]

            if triggered_id == "close":
                return False, dash.no_update
            else:
                if triggered_id == "load_expe":
                    metabo_exp_dto = decode_pickle_from_base64(contents_loaded)
                    if self.metabo_controller.is_save_safe(metabo_exp_dto):
                        self.metabo_controller.full_restore(metabo_exp_dto)
                        return (
                            False,
                            "reload",
                            "",
                        )
                    else:
                        return True, dash.no_update, dash.no_update

                elif triggered_id == "loadAnyway":
                    metabo_exp_dto = decode_pickle_from_base64(contents_loaded)
                    self.metabo_controller.load_results(metabo_exp_dto)
                    return (
                        False,
                        "",
                        "",
                    )  # dcc.Location(href="/home", id="someid_doesnt_matter")

                elif triggered_id == "partialRestore":
                    metabo_exp_dto = decode_pickle_from_base64(contents_loaded)
                    if new_data and new_metadata:
                        try:
                            self.metabo_controller.partial_restore(
                                metabo_exp_dto,
                                new_data_name,
                                new_metadata_name,
                                data=new_data,
                                metadata=new_metadata,
                            )
                        except ValueError as ve:
                            return True, dash.no_update, str(ve)
                        return (
                            False,
                            "",
                            "",
                        )
                    else:
                        return (
                            True,
                            "",
                            "You need to upload both data and metadata to do a partial restore",
                        )

                elif triggered_id == "fullRestore":
                    metabo_exp_dto = decode_pickle_from_base64(contents_loaded)
                    if (
                        new_data is not None
                        and new_metadata is not None
                        and Utils.are_files_corresponding_to_dto(
                            new_data, new_metadata, metabo_exp_dto
                        )
                    ):
                        self.metabo_controller.full_restore(metabo_exp_dto)
                        return (
                            False,
                            "",
                            "",
                        )
                    else:
                        return (
                            True,
                            "",
                            "You need to restore original data matrix and metadata to do a full restore",
                        )

                return False, dash.no_update, dash.no_update



# Number of features to compute PCA, UMAP and Strip charts
features = [5, 10, 40, 100]

# Default marks for sliders
default_marks = {str(k): str(v) for k, v in enumerate(features)}

# Mark for the all value
all_mark = {str(len(features)): "All"}

# Offset when showing the used value
custom_mark_offset = 0.3

# Limit of used features to plot (used for strip charts)
max_used_features_to_show = 200


import os
import pickle
from typing import Tuple, Iterable, Union

import pandas as pd
from sklearn.preprocessing import StandardScaler

from ..service import DataFormat, compute_hash, Utils

ROOT_PATH = os.path.dirname(__file__)
DUMP_PATH = os.path.join(ROOT_PATH, "dumps")
DUMP_DATA_MATRIX_PATH = os.path.join(DUMP_PATH, "datamatrix.p")


class DataMatrix:
    def __init__(self):
        # TODO : implémenter test format de matrice (progenesis -> ML ready)
        self.data = None
        self._scaler = StandardScaler()
        self._hash = None
        self._use_raw = None
        self._remove_rt = True

    def reset_file(self):
        """
        Reset ? the content of the DUMP_DATA_MATRIX_PATH file
        """
        Utils.reset_file(DUMP_DATA_MATRIX_PATH)

    def read_format_and_store_data(self, path: str, data=None, from_base64: bool = True,) -> tuple[pd.DataFrame|None, list[str]]:
        """
        As the name states,
        Read a file of data, format what needs to be, and store it/return it
        """
        if self._use_raw is None:
            raise RuntimeError("Need to set raw use before loading data")
        # metadata_df will be None in the case of a "general" file, and will contain something in case of
        # Progenesis file (an artificial metadata dataframe is created in that case). 
        data_df, metadata_df = self._load_and_format(path, data=data, is_raw=self._use_raw, from_base64=from_base64)

        if self._remove_rt:
            features_to_drop = []
            for f in data_df.columns:
                if float(f.split("_")[0]) < 1:
                    features_to_drop.append(f)
            data_df = data_df.drop(columns=features_to_drop)

        if data is not None:
            self._hash = compute_hash(data)

        with open(DUMP_DATA_MATRIX_PATH, "w+b") as data_matrix_file:
            pickle.dump(data_df, data_matrix_file)

        self._scaler.fit(data_df)
        
        # unique_ids will contain sample names from the data file
        unique_ids: list[str] = data_df.index.to_list() if metadata_df is None else metadata_df["sample_names"].tolist()
        return metadata_df, unique_ids
        # The event of having metadata_df=None is handled in the set_data_matrix function that calls this function
        # So no need to add an "else" here

    def get_hash(self) -> str:
        """
        :return: the hash of the dataframe
        """
        return self._hash

    def is_raw(self) -> Union[bool, None]:
        """
        return the bool indicating to use either raw or normalized data
        """
        return self._use_raw

    def get_scaled_data(self, selected_ids: Iterable = None) -> pd.DataFrame:
        """
        Scale the dataframe to be ready for ML
        :param data: the dataframe to scale
        :param selected_ids: the list of IDs to scale
        :return: the scaled dataframe
        """
        if self.data is None:
            raise RuntimeError("Need to load data from file before scaling")
        if selected_ids is not None:
            df = self.data.loc[selected_ids, :]
        else:
            df = self.data
        return pd.DataFrame(self._scaler.transform(df), columns=self.data.columns)

    def _load_and_format(self, path, data=None, is_raw=False, from_base64=True) -> Tuple[pd.DataFrame, Union[pd.DataFrame, None]]:
        """
        load the table from a path and process it to make it more easy to manipulate
        """
        formater = DataFormat(path, data=data, use_raw=is_raw, from_base64_str=from_base64)
        datatable_compoundsInfo, datatable, labels, sample_names = formater.convert()
        if labels is None or sample_names is None:
            # This if triggers (for now) when a "general" file is read
            return datatable, None
        # This return triggers when a Progenesis file is read
        return datatable, pd.DataFrame({"sample_names": sample_names, "labels": labels})

    def load_data(self):  # loadDataMatrix
        with open(DUMP_DATA_MATRIX_PATH, "rb") as data_matrix_file:
            self.data = pickle.load(data_matrix_file)

    def load_samples_corresponding_to_IDs_in_splits(
        self, id_list: list
    ) -> pd.DataFrame:
        """
        self.data is supposed to have samples as lines, so we select only the samples we need
        with their IDS(lines indexes)
        Also, the way of selecting the samples assures the samples are in the same order as the list, which is
        it self in the same order as the target list
        :return: the reduced dataframe, the samples are lines and the matrix is ML ready
        """
        if self.data is None:
            raise RuntimeError(
                "Need to load data from file before extracting specific samples"
            )
        df = self.data.loc[id_list, :]
        return df

    def unload_data(self):
        self.data = None

    def set_raw_use(self, use_raw: bool):
        """
        Set the bool value of whether to use the raw data from a progenesis matrix or the normalized data
        """
        self._use_raw = use_raw

    def set_remove_rt(self, remove_rt: bool):
        """
        Set the bool value of whether to remove the features detected before 1minRT (Retention Time)
        (Those features are usually noise and should not be considered as actual metabolites/features)
        """
        self._remove_rt = remove_rt

    def get_remove_rt(self) -> bool:
        """
        return the value of the _remove_rt attribute
        (if true, remove the features detected before 1 minute of Retention Time)
        """
        return self._remove_rt

    def data_is_set(self) -> bool:
        return self._hash is not None
  


import json
import re

import dash_bootstrap_components as dbc
import pandas as pd
from dash import html, State, Input, Output, dash, dcc, callback_context, ALL
from os.path import basename

from .MetaTab import MetaTab
from ...service import Utils, init_logger, log_exceptions
from ...domain import Controller


class MLTab(MetaTab):
    def __init__(self, app: dash.Dash, metabo_controller: Controller):
        super().__init__(app, metabo_controller)
        self._logger = init_logger()

    def getLayout(self) -> dbc.Tab:       
        __splitConfigFile = html.Div(
            [
                dbc.Label("Select CV search type", className="form_labels"),
                dbc.RadioItems(
                    options=Utils.format_list_for_checklist(
                        self.metabo_controller.get_cv_types()
                    ),
                    value=self.metabo_controller.get_selected_cv_type(),
                    id="radio_cv_types",
                ),
                html.Br(),
                html.Div(id="cv_params"),
            ],
            className="form_field",
        )

        __ThreadingConfig = html.Div(
            [
                dbc.Label("Multi-threading activation", className="form_labels"),
                dbc.Checklist(
                    options=[
                        {"label": "Multi-threading is off", "value": 1},
                    ],
                    id="switches-input",
                    value=[1],
                    switch=True,
                ),
                html.Div(id="output_cv_folds_ml", style={"color": "red"}),
            ],
            className="form_field",
        )

        _definitionLearningConfig = html.Div(
            className="title_and_form",
            children=[
                html.H4(id="Learn_conf_title", children="Define Learning configs"),
                dbc.Form(
                    children=[
                        dbc.Col(
                            children=[
                                __splitConfigFile,
                                __ThreadingConfig,
                            ],
                        )
                    ]
                ),
            ],
        )

        __availableAlgorithms = html.Div(
            [
                dbc.Label("Available Algorithms", className="form_labels"),
                dbc.Checklist(
                    id="in_algo_ML",
                    # inline=True
                ),
                html.Div(id="output_checklist_ml", children="", style={"color": "red"}),
            ],
            className="form_field",
        )

        __addCustomAlgorithm = html.Div(
            [
                dbc.Label("Add Sklearn Algorithms", className="form_labels"),
                dbc.Label("from sklearn.A import B"),
                dbc.Input(
                    id="import_new_algo",
                    placeholder="Complete import (A)",
                    className="form_input_text",
                ),
                dbc.Input(
                    id="name_new_algo",
                    placeholder="Enter Name (B)",
                    className="form_input_text",
                ),
                dbc.Button(
                    "Get attributes",
                    color="success",
                    id="get_attribute_button",
                    className="custom_buttons",
                    n_clicks=0,
                ),
                dbc.Button(
                    "Manual configuration",
                    color="success",
                    id="manual_config_button",
                    className="custom_buttons",
                    n_clicks=0,
                ),
                html.Div(id="output_import_algo"),
                html.Div("WARNING: Incorrect configuration may lead to errors", style={"color": "orange"}),
                html.Br(),
                dbc.Label("Specify parameters to explore by gridsearch"),
                html.Div(id="table_param"),
                html.Div(id="output_error_import_algo"),
                dbc.Label("Specify importance attribute"),
                dbc.Select(
                    id="importance_attributes_dropdown_menu",
                ),
                dbc.Button(
                    "Add",
                    color="success",
                    id="add_n_refresh_sklearn_algo_button",
                    className="custom_buttons",
                    n_clicks=0,
                ),
            ],
            className="form_field",

        )

        __validationButton = html.Div(
            id="Learning_button_box",
            className="button_box",
            children=[
                dcc.Loading(
                    id="learn_loading",
                    children=[
                        html.Div(
                            id="learn_loading_output",
                            style={"color": "green"},
                        )
                    ],
                    style={"width": "100%"},
                    type="dot",
                    color="#13BD00",
                ),
                dbc.Button(
                    "Learn",
                    color="primary",
                    id="start_learning_button",
                    className="custom_buttons",
                    n_clicks=0,
                ),
                html.Div(id="output_button_ml", children="", style={"display": "none"}),
            ], )

        _definitionLearningAlgorithm = html.Div(
            className="title_and_form",
            children=[
                html.H4(id="learn_algo_title", children="Define Learning Algorithms"),
                dbc.Form(
                    children=[
                        dbc.Col(
                            children=[
                                __availableAlgorithms,
                                # TODO: change button style
                                dbc.Button(
                                    "Add custom sklearn algorithm (for experts)",
                                    id="collapse-button",
                                    className="mb-3",
                                    color="outline-primary",
                                    n_clicks=0,
                                ),
                                dbc.Collapse(
                                    dbc.Card(dbc.CardBody(__addCustomAlgorithm)),
                                    id="collapse",
                                    is_open=False,
                                ),
                                __validationButton,
                            ],
                        )
                    ]
                ),
            ],
        )

        _learn_completed_modal = dbc.Modal(
            [
                dbc.ModalHeader(dbc.ModalTitle("Learning completed")),
                dbc.ModalBody(id="learn_completed_body", children="The learning is completed."),
                dbc.ModalFooter(
                    dbc.Button("Close", id="close-learn_completed_modal", className="ml-auto")
                ),
            ],
            id="learn_completed_modal",
            is_open=False,
            centered=True
        )

        return dbc.Tab(
            className="global_tab",
            label="Machine Learning",
            children=[
                dcc.Store(id='learn-button-state', data=False),
                dcc.Store(id='classification_design_filename-store', data=""),
                html.Div(
                    className="fig_group",
                    children=[_definitionLearningConfig, _definitionLearningAlgorithm, _learn_completed_modal],
                ),
            ],
        )

    def _registerCallbacks(self) -> None:
        @self.app.callback(
            Output("in_nbr_CV_folds", "value"), [Input("custom_big_tabs", "active_tab")]
        )
        @log_exceptions(self._logger)
        def update_nbr_CV_folds(active_tab):
            if active_tab == "tab-2":
                return self.metabo_controller.get_cv_folds()
            return dash.no_update

        @self.app.callback(
            Output("switches-input", "options"), [Input("switches-input", "value")]
        )
        def update_switch(value):
            if value == [1]:
                self.metabo_controller.set_multithreading(True)
                return [
                    {"label": "Multi-threading is on", "value": 1},
                ]
            elif not value:
                self.metabo_controller.set_multithreading(False)
                return [
                    {"label": "Multi-threading is off", "value": 1},
                ]
            return dash.no_update

        @self.app.callback(
            [
                Output("in_algo_ML", "options"),
                Output("in_algo_ML", "value"),
                Output("output_checklist_ml", "children"),
                Output("output_error_import_algo", "children"),
            ],
            [
                Input("add_n_refresh_sklearn_algo_button", "n_clicks"),
                Input("in_algo_ML", "value"),
                Input("custom_big_tabs", "active_tab"),
            ],
            [
                State("import_new_algo", "value"),
                State("name_new_algo", "value"),
                State("table_param", "children"),
                State("importance_attributes_dropdown_menu", "value")
            ]
        )
        @log_exceptions(self._logger)
        def add_refresh_available_sklearn_algorithms(
                n, value, active_tab, import_algo, name_algo, table_param, importance_attribute
        ):
            triggered_by = callback_context.triggered[0]["prop_id"].split(".")[0]
            if triggered_by == "custom_big_tabs":
                self._logger.info("Triggered by tab")
                if active_tab == "tab-2":
                    self.metabo_controller.update_classification_designs_with_selected_models()

            if triggered_by == "add_n_refresh_sklearn_algo_button":
                grid_search_params = {}
                model = Utils.get_model_from_import([import_algo], name_algo)
                param_types = {param: type for param, type in Utils.get_model_parameters(model)}
                error_children = []
                if "{'props': {'children': 'Name', 'colSpan': 1}, 'type': 'Th', 'namespace': 'dash_html_components'}" in str(
                        table_param[-1]["props"]):
                    params_and_values = re.findall(r"{'id': '(\w+)'[\w,' :]*'value': '([\[\],. \w]+)',[\w,': ]*}",
                                                   str(table_param))
                    try:
                        grid_search_params = Utils.transform_params_to_cross_validation_dict(params_and_values,
                                                                                             param_types)
                    except ValueError as e:
                        for error in e.args[0]:
                            error_children.append(html.P(error, style={"color": "red"}))
                else:
                    json_params_and_values = table_param[-1]["props"]["value"]
                    try:
                        params_and_values = json.loads(json_params_and_values)
                    except json.decoder.JSONDecodeError as e:
                        return dash.no_update, dash.no_update, dash.no_update, html.P("Error in parameters:" + e.msg,
                                                                                      style={"color": "red"})

                if importance_attribute is None:
                    error_children.append(html.P("Please select an importance attribute", style={"color": "red"}))

                if error_children:
                    return dash.no_update, dash.no_update, dash.no_update, error_children

                self.metabo_controller.add_custom_model(
                    name_algo, import_algo, grid_search_params, importance_attribute
                )
            if triggered_by == "in_algo_ML":
                self._logger.info("Triggered by dropdown")
                try:
                    self.metabo_controller.set_selected_models(value)
                except ValueError as ve:
                    self._logger.error(f"{ve}")
                    return (
                        Utils.format_list_for_checklist(
                            self.metabo_controller.get_all_algos_names()
                        ),
                        [],
                        str(ve),
                        "",
                    )

            return (
                Utils.format_list_for_checklist(
                    self.metabo_controller.get_all_algos_names()
                ),
                self.metabo_controller.get_selected_models(),
                "",
                "",
            )

        @self.app.callback(
            [
                Output("output_button_ml", "children"),
                Output("learn_loading_output", "children"),
                Output("learn-button-state", "data"),
                Output('classification_design_filename-store', 'data'),
            ],
            [Input("start_learning_button", "n_clicks")],
        )
        @log_exceptions(self._logger)
        def start_machine_learning(n):
            if n >= 1:
                self._logger.info(f"in\n{self.metabo_controller.get_selected_models()}")
                self.metabo_controller.learn()

                # Dump file to dump folder and to save folder (backup)
                metabo_expe_filename = Utils.get_metabo_experiment_path("pico_ml")
                metabo_expe_obj = self.metabo_controller.generate_save()
                Utils.dump_metabo_expe(metabo_expe_obj) # Dump the classification design to the dump folder
                Utils.dump_metabo_expe(metabo_expe_obj, metabo_expe_filename) # Save classification design.
                del metabo_expe_obj
                self._logger.info(f'The classification design file "{metabo_expe_filename}" has been saved.')
                self._logger.info(f"bip bip 3 : {self.metabo_controller._metabo_experiment.classification_designs}")
                return "Done!", "", True, basename(metabo_expe_filename)
            else:
                return dash.no_update

        @self.app.callback(
            Output("learn_completed_modal", "is_open"),
            [Input("close-learn_completed_modal", "n_clicks"),
            Input('learn-button-state', 'data')],
            [State("learn_completed_modal", "is_open")]
        )
        def toggle_learn_completed_modal(close_clicks, learn_state, is_open):
            if close_clicks or learn_state:
                return not is_open
            return is_open

        @self.app.callback(
        Output("learn_completed_body", "children"),
        [Input('classification_design_filename-store', 'data')]
        )
        @log_exceptions(self._logger)
        def update_modal_message(filename):
            if filename:
                return [
                    html.P(f'The classification design file "{filename}" has been saved.', style={'color': 'green'}),
                    html.P('You can go to the "Results" tabs.'),
                ]
            return "Error!"

        @self.app.callback(
            [Output("radio_cv_types", "value"),
             Output("cv_params", "children")],
            [Input("radio_cv_types", "value"),
             Input("custom_big_tabs", "active_tab"),
             Input({"type": "cv_params", "index": ALL}, "value")],
        )
        @log_exceptions(self._logger)
        def set_cv_type(cv_value, tab, input_params):
            if callback_context.triggered[0]["prop_id"] == "radio_cv_types.value":
                params_form = []
                self.metabo_controller.set_cv_type(cv_value)
                params = self.metabo_controller.get_cv_algorithm_configuration()
                if params:
                    for param in params:
                        if not param["constant"]:
                            name = param["name"]
                            value = param["value"]
                            type = param["type"]
                            if type == "int":
                                html_type = "number"
                            elif type == "float":
                                html_type = "number"
                            elif type == "bool":
                                html_type = "checkbox"
                            else:
                                html_type = "text"
                            params_form.append(
                                html.Tr(
                                    [
                                        html.Td(name),
                                        html.Td(
                                            dcc.Input(
                                                id={"type": "cv_params", "index": name},
                                                type=html_type,
                                                value=value,
                                            )
                                        ),
                                    ]
                                )
                            )
                return cv_value, params_form
            else:
                if input_params is not None and input_params != []:
                    self.metabo_controller.set_cv_algorithm_configuration(input_params)

            return self.metabo_controller.get_selected_cv_type(), dash.no_update

        @self.app.callback(
            [Output("output_import_algo", "children"),
             Output("output_import_algo", "style"),
             Output("table_param", "children"),
             Output("importance_attributes_dropdown_menu", "options")],
            [Input("get_attribute_button", "n_clicks"),
             Input("manual_config_button", "n_clicks")],
            [State("import_new_algo", "value"),
             State("name_new_algo", "value")],
        )
        @log_exceptions(self._logger)
        def get_attribute_algo(n_attribute, n_manual, import_new, new_algo_name):
            triggered_by = callback_context.triggered[0]["prop_id"].split(".")[0]
            self._logger.info(f"triggered by: [{triggered_by}]")
            if n_manual > 0 or n_attribute > 0:
                try:
                    model = Utils.get_model_from_import([import_new], new_algo_name)
                except Exception as e:
                    self._logger.error(f"{e}")
                    return "Import failed: " + str(e), {"color": "red"}, "", ""
                importance_attributes = [param_name for param_name, _ in
                                         Utils.get_model_parameters_after_training(model)]
                if not importance_attributes:
                    return "Import failed: No importance attribute found.", {"color": "red"}, "", ""
                if triggered_by == "get_attribute_button":
                    try:
                        attributes = Utils.get_model_parameters(model)
                        attributes_table = pd.DataFrame(attributes, columns=["Name", "Type"])
                        attributes_table["Type"].replace(
                            {"str": "String", "int": "Integer", "float": "Float", "NoneType": "Unspecified"},
                            inplace=True)
                        inputs = []
                        for attribute, _ in attributes:
                            inputs.append(dbc.Input(id=attribute, type="text", placeholder="Value"))
                        attributes_table["Value"] = inputs

                        default_text = [
                            html.Br(),
                            html.P("You can set the grid search parameters as followed:"),
                            html.P(
                                "Values: 'val1A, val1B, val1C'"
                            ),
                            dbc.Table.from_dataframe(attributes_table)
                        ]

                        return f"{model.__name__} found", {
                            "color": "green"}, default_text, Utils.format_list_for_checklist(importance_attributes)
                    except Exception as e:
                        self._logger.info(f"{e}")
                        return "Import failed: " + str(e), {"color": "red"}, "", ""
                elif triggered_by == "manual_config_button":
                    return "", None, [html.P("The following configuration must be in JSON format",
                                             style={"color": "orange"}), html.Br(),
                                      dcc.Textarea()], Utils.format_list_for_checklist(importance_attributes)
            return dash.no_update, dash.no_update, dash.no_update, dash.no_update

        @self.app.callback(
            Output("collapse", "is_open"),
            [Input("collapse-button", "n_clicks")],
            [State("collapse", "is_open")],
        )
        def toggle_collapse(n, is_open):
            if n:
                return not is_open
            return is_open



import dash_bootstrap_components as dbc
from dash import html

from .MetaTab import MetaTab


class InterpretTab(MetaTab):
    def getLayout(self) -> dbc.Tab:
        return dbc.Tab(
            className="global_tab",
            # tab_style={"margin-left": "auto"},
            label="Model Interpretation",
            children=[
                html.P(
                    "A tab to allow model interpretation with model-agnostic methods."
                )
            ],
        )

    def _registerCallbacks(self) -> None:
        pass



from collections import Counter

import dash_bootstrap_components as dbc
import numpy as np
from dash import html, dcc, Output, Input, State, dash, Dash

from .MetaTab import MetaTab
from ...service import Plots, init_logger, log_exceptions
from ...domain import Controller

CONFIG = {
    "toImageButtonOptions": {
        "format": "svg",  # one of png, svg, jpeg, webp
        "height": None,
        "width": None,
        "scale": 1,  # Multiply title/legend/axis/canvas sizes by this factor
    }
}


class AggregatedResultsTab(MetaTab):
    def __init__(self, app: Dash, metabo_controller: Controller):
        super().__init__(app, metabo_controller)
        self._logger = init_logger()
        self.r = self.metabo_controller.get_all_results()
        self._plots = Plots("blues")

    def getLayout(self) -> dbc.Tab:
        _resultsMenuDropdowns = dbc.Card(
            className="results_menu_dropdowns",
            children=[
                dbc.CardBody(
                    [
                        html.Div(
                            className="dropdowns",
                            children=[
                                html.H6("Classification Design : "),
                                dbc.Select(
                                    id="design_dropdown_summary",
                                    className="form_select",
                                    options=[{"label": "None", "value": "None"}],
                                    value="None",
                                ),
                                dcc.Store(id='design_dropdown_summary_store', storage_type='session'),
                            ],
                        ),
                        dbc.Button(
                            "Load",
                            color="primary",
                            id="load_results_button",
                            className="custom_buttons",
                            n_clicks=0,
                        ),
                        html.Div(id="output_button_load_results"),
                    ],
                    id="all_algo_results",
                )
            ],
        )
        
        __currentExperimentInfo = dbc.Card(
            children=[
                dbc.CardBody(
                    children=[
                        html.H6(
                            "Current experiment info"
                        ),  
                        dcc.Loading(
                            id="loading_expe_table_summary",
                            children=html.Div(id="expe_table_summary", children=""),
                            type="circle",
                        ),
                    ]
                )
            ],
            className="w-25",
        )

        _resultsInfo = html.Div(
            className="Results_info",
            children=[
                _resultsMenuDropdowns,
                __currentExperimentInfo,
            ],
        )


        _heatmapUsedFeatures = html.Div(
            className="umap_plot_and_title",
            children=[
                html.Div(
                    className="title_and_help",
                    children=[
                        html.H6("Features Usage"),
                        dbc.Button(
                            "[?]",
                            className="text-muted btn-secondary popover_btn",
                            id="help_heatmapFeatures",
                        ),
                        dbc.Popover(
                            children=[dbc.PopoverBody("Blablabla wout wout")],
                            id="pop_help_heatmapFeatures",
                            is_open=False,
                            target="help_hHeatmapFeatures",
                        ),
                    ],
                ),
                dcc.Loading(
                    dcc.Graph(id="heatmapFeatures", config=CONFIG),
                    type="dot",
                    color="#13BD00",
                ),
                dcc.Slider(-3, 0, 0.1,
                           id='slider_heatmapFeatures',
                           marks={i: '{}'.format(10 ** i) for i in range(-3, 1, 1)},
                           value=-2,
                           updatemode='drag',
                           ),
            ],
        )

        _heatmapSamplesAlwaysWrong = html.Div(
            className="umap_plot_and_title",
            children=[
                html.Div(
                    className="title_and_help",
                    children=[
                        html.H6("Errors on samples in test"),
                        dbc.Button(
                            "[?]",
                            className="text-muted btn-secondary popover_btn",
                            id="help_heatmapSamples",
                        ),
                        dbc.Popover(
                            children=[dbc.PopoverBody("Blablabla wout wout")],
                            id="pop_help_heatmapSamples",
                            is_open=False,
                            target="help_heatmapSamples",
                        ),
                    ],
                ),
                dcc.Loading(
                    dcc.Graph(id="heatmapSamples", config=CONFIG),
                    type="dot",
                    color="#13BD00",
                ),
                # dcc.Slider(min=0, max=3, step=1, value=0, marks={0: "10", 1: "40", 2: "100", 3: "All"},
                #            id="features_stripChart_dropdown")
            ],
        )

        _barplotComparaisonAlgo = html.Div(
            className="umap_plot_and_title",
            children=[
                html.Div(
                    className="title_and_help",
                    children=[
                        html.H6("Comparaison of algorithms performances"),
                        dbc.Button(
                            "[?]",
                            className="text-muted btn-secondary popover_btn",
                            id="help_barplotAlgo",
                        ),
                        dbc.Popover(
                            children=[dbc.PopoverBody("Blablabla wout wout")],
                            id="pop_help_barplotAlgo",
                            is_open=False,
                            target="help_barplotAlgo",
                        ),
                    ],
                ),
                dcc.Loading(
                    dcc.Graph(id="barplotAlgo", config=CONFIG),
                    type="dot",
                    color="#13BD00",
                ),
            ],
        )

        return dbc.Tab(
            className="global_tab",
            label="Results aggregated",
            children=[
                _resultsInfo,
                html.Div(
                    className="fig_group",
                    children=[
                        _barplotComparaisonAlgo
                    ],
                ),
                html.Div(className="fig_group", children=[
                    _heatmapSamplesAlwaysWrong,
                    _heatmapUsedFeatures
                ]),
            ],
        )
        # html.Div(className="column_content",
        #          # WARNING !! : _infoFigure is not with the card, it's in a separate column
        #          children=[_heatmapSamplesAlwaysWrong])])])

    def _registerCallbacks(self) -> None:
        @self.app.callback(
            [
                Output("design_dropdown_summary", "options"),
                Output("design_dropdown_summary", "value"),
            ],
            Input("custom_big_tabs", "active_tab"),
            State('design_dropdown_summary_store', 'data'),
        )
        @log_exceptions(self._logger)
        def update_results_dropdown_design(active, stored_value):
            if active == "tab-4":
                try:
                    self.r = self.metabo_controller.get_all_results()
                    a = list(self.r.keys())
                    if stored_value is not None and stored_value in a:
                        return [{"label": i, "value": i} for i in a], stored_value
                    return [{"label": i, "value": i} for i in a], a[0]
                except:  # TODO: wrong practice ???
                    return dash.no_update, dash.no_update
            else:
                return dash.no_update, dash.no_update

        @self.app.callback(
            Output('design_dropdown_summary_store', 'data'),
            Input('load_results_button', 'n_clicks'),
            State('design_dropdown_summary', 'value')
        )
        def save_design_dropdown_summary_value(_, value):
            return value
        
        @self.app.callback(
            [
                Output("expe_table_summary", "children")
            ],
            [
                Input("load_results_button", "n_clicks")
            ],
            [
                State("design_dropdown_summary", "value")
            ],
        )
        @log_exceptions(self._logger)
        def get_experiment_statistics(_, design_name):
            if design_name == "None":
                return dash.no_update
            
            df = self.r[design_name][list(self.r[design_name].keys())[0]].results["info_expe"]
            table_body = self._plots.show_exp_info_all(df)
            table = dbc.Table(
                table_body, id="table_exp_info", borderless=True, hover=True
            )
            return [table]

        @self.app.callback(
            [
                Output("heatmapFeatures", "figure"),
            ],
            [Input("load_results_button", "n_clicks"),
             Input("slider_heatmapFeatures", "value")],
            [State("design_dropdown_summary", "value"),
             ],
        )
        @log_exceptions(self._logger)
        def show_heatmap_features_usage(n_clicks, log_importance_threshold, design):
            if n_clicks >= 1:

                importance_threshold = 10 ** log_importance_threshold

                algos = list(self.r[design].keys())
                global_df = None
                for a in algos:
                    if global_df is None:
                        self._logger.info("Global df is none")
                        global_df = self.r[design][a].results["features_table"]
                        global_df = global_df.loc[
                            :, ("features", "importance_usage")
                        ]  # reduce dataframe to 2 columns
                        global_df.rename(
                            columns={"importance_usage": a}, inplace=True
                        )  # rename column to identify algorithm
                    else:
                        self._logger.info(f"Global df not none, algo : {a}")
                        df = self.r[design][a].results[
                            "features_table"
                        ]  # retrieve features table of algo a
                        df = df.loc[
                            :, ("features", "importance_usage")
                        ]  # reduce dataframe to 2 columns
                        df.rename(
                            columns={"importance_usage": a}, inplace=True
                        )  # rename column to identify algorithm
                        global_df = global_df.merge(
                            df, how="outer", on="features"
                        )  # join data with global dataset

                global_df = global_df.set_index("features")
                global_df = global_df.fillna(0)

                global_df = global_df.loc[global_df.max(axis=1) >= importance_threshold]

                hm_fig = self._plots.show_heatmap_features_usage(global_df, importance_threshold)

                return [hm_fig]
            else:
                return dash.no_update

        @self.app.callback(
            Output("heatmapSamples", "figure"),
            [Input("load_results_button", "n_clicks")],
            State("design_dropdown_summary", "value"),
        )
        @log_exceptions(self._logger)
        def show_heatmap_samples_always_wrong(n_clicks, design):
            if n_clicks >= 1:
                algos = list(self.r[design].keys())

                data_train = []
                data_test = []
                all_samples = []

                for i, a in enumerate(algos):
                    data_train.append([])
                    data_test.append([])
                    train = []
                    test = []
                    for j, s in enumerate(self.r[design][a].splits_number):
                        train_d, test_d = self.r[design][a].results[s]["failed_samples"]
                        train.append(train_d)
                        test.append(test_d)

                    counter_train = Counter()
                    for d in train:
                        counter_train.update(d)

                    counter_test = Counter()
                    for d in test:
                        counter_test.update(d)

                    all_samples = list(counter_train.keys()) + list(counter_test.keys())
                    for s in all_samples:
                        if s in counter_train.keys():
                            data_train[i].append(counter_train[s])
                        else:
                            data_train[i].append(0)

                        if s in counter_test.keys():
                            data_test[i].append(counter_test[s])
                        else:
                            data_test[i].append(0)

                data_train = np.array(data_train).T
                data_test = np.array(data_test).T

                fig = self._plots.show_heatmap_wrong_samples(
                    data_train, data_test, all_samples, algos
                )

                return fig
            else:
                return dash.no_update

        @self.app.callback(
            Output("barplotAlgo", "figure"),
            [Input("load_results_button", "n_clicks")],
            State("design_dropdown_summary", "value"),
        )
        @log_exceptions(self._logger)
        def show_barplot_compare_accuracy_algo(n_clicks, design_name):
            """
            retrieve balanced accuracy values now
            """
            if n_clicks >= 1:
                algos = list(self.r[design_name].keys())

                train_acc = []
                train_std = []
                test_acc = []
                test_std = []
                for a in algos:
                    df = self.r[design_name][a].results["metrics_table"]
                    train_m, train_s = df["train"][1].split("(")
                    train_acc.append(float(train_m))
                    train_std.append(float(train_s.split(")")[0]))

                    test_m, test_s = df["test"][1].split("(")
                    test_acc.append(float(test_m))
                    test_std.append(float(test_s.split(")")[0]))

                fig = self._plots.show_barplot_comparaison_algo(algos, train_acc, train_std, test_acc, test_std)
                return fig
            else:
                return dash.no_update




from sklearn.model_selection import GridSearchCV, RandomizedSearchCV

CV_ALGORITHMS = {
    "GridSearchCV": {
        "constructor": GridSearchCV,
        "params": []
    }, 
    "RandomizedSearchCV": {
        "constructor": RandomizedSearchCV,
        "params": [{
            "name": "n_iter", 
            "value": 10, 
            "type": "int", 
            "constant": False
        }]
    }
}



import base64
import hashlib
import importlib
import os
import pickle
from typing import Union
import re

import pickle as pkl
from typing import List, Dict, Tuple

import numpy
import numpy as np
import pandas as pd
import sklearn

PACKAGE_ROOT_PATH = os.sep.join(os.path.dirname(__file__).split(os.sep)[:-1])
DUMP_PATH = os.path.join(PACKAGE_ROOT_PATH, "domain", "dumps")
DUMP_EXPE_PATH = os.path.join(DUMP_PATH, "save.mtxp")

DEFAULT_IMPORTANCE_ATTRIBUTE = "feature_importances_"


def dump_metabo_expe(obj, expe_file_path: str=DUMP_EXPE_PATH):
    with open(expe_file_path, "w+b") as expe_file:
        pkl.dump(obj, expe_file)

        
def get_dumped_metabo_experiment_path() -> str:
    return DUMP_EXPE_PATH


def get_metabo_experiment_path(expe_filename: str="save", is_date: bool=True) -> str:
    """Get the metabo experiment path.
        - optionaly add the date (if is_date is True).
        - add the ".mtxp" extension if expe_filename don't have one.

    Args:
        expe_filename (str, optional): filename to use as template. Defaults to "save".
        is_date (bool, optional): True to add the date and time in the filename. Defaults to True.

    Returns:
        str: The full path to the filename
    """
    _, name_ext = os.path.splitext(expe_filename)
    if not name_ext:
        expe_filename = expe_filename + ".mtxp"
    
    saves_dir = get_pico_subdir("saves")
    
    new_filename = insert_datetime(expe_filename) if is_date else expe_filename
    return os.path.join(saves_dir, new_filename)


def get_pico_subdir(subdir_name: str) -> str:
    """Returns the path to provided subdirectory of '~/pico_files/' main pico directory.

    Args:
        subdir_name (str): subdir to get (and create if it doesn't exist)

    Returns:
        str: the subdir full path
    """
    home_dir_path: str = os.path.expanduser("~")
    subdir_path: str = os.path.join(home_dir_path, "pico_files", subdir_name)
    
    if not os.path.exists(subdir_path):
        os.makedirs(subdir_path)
        
    if not os.path.exists(subdir_path):
        subdir_path = os.getcwd()
    
    return subdir_path


from datetime import datetime

def insert_datetime(expe_filename: str) -> str:
    current_datetime = datetime.now().strftime("_%Y%m%d_%H%M%S")
    name_root, name_ext = os.path.splitext(expe_filename)
    return name_root + current_datetime + name_ext


def load_metabo_expe(path):
    if os.path.isfile(DUMP_EXPE_PATH):
        with open(path, "rb") as expe_file:
            return pkl.load(expe_file)
    else:
        return None


def retrieve_data_from_sample_name(names_list, dataframe):
    """

    :param names_list: list of samples name
    :param dataframe: a dataframe with each sample as a line identified by the samples' name
    :return: list of data
    """
    print("retrieving data from name")
    data_list = []
    for n in names_list:
        d = dataframe.loc[n, :]
        data_list.append(d.tolist())
    # print("data list : {}".format(data_list[0]))
    print("data from name retrieved")
    # print("data 2nd element : {}".format(data_list[1]))
    return data_list


def list_filler(liste):
    """
    (NA or "")
    Complete the NA values of a list with the last non NA value from left to right.
    If the first value is NA, then leave it like this.
    :param liste: list to fill
    :return: new list filled
    """
    l = []
    current = ""
    for idx, j in enumerate(liste):
        if j != "":
            current = j
        l.append(current)
    return l


def read_Progenesis_compounds_table(fileName, with_raw=True):
    datatable = pd.read_csv(fileName, header=2, index_col=0)
    header = pd.read_csv(fileName, nrows=1, index_col=0)
    start_normalized = header.columns.tolist().index("Normalised abundance")

    labels_array = np.array(header.iloc[0].tolist())
    possible_labels = labels_array[labels_array != "nan"]

    if with_raw:
        start_raw = header.columns.tolist().index("Raw abundance")
        sample_names = datatable.iloc[:, start_normalized:start_raw].columns
        possible_labels = possible_labels[0: int(len(possible_labels) / 2)]
    else:
        sample_names = datatable.iloc[:, start_normalized:].columns

    labels = [""] * len(sample_names)
    start_label = possible_labels[0]
    labels_array = labels_array.tolist()
    for next_labels in possible_labels[1:]:
        index_s = labels_array.index(start_label) - start_normalized
        index_e = labels_array.index(next_labels) - start_normalized
        labels[index_s:index_e] = [start_label] * (index_e - index_s)
        start_label = next_labels
    labels[index_e:] = [start_label] * (len(labels) - index_e)

    labels_dict = {sample_names[i]: j for i, j in enumerate(labels)}

    if with_raw:
        datatable_compoundsInfo = datatable.iloc[:, 0:start_normalized]
        datatable_normalized = datatable.iloc[:, start_normalized:start_raw]
        datatable_raw = datatable.iloc[:, start_raw:]
        datatable_raw.columns = [
            i.rstrip(".1") for i in datatable_raw.columns
        ]  # Fix the columns names

        datatable_normalized = datatable_normalized.T
        datatable_raw = datatable_raw.T
        datatable_compoundsInfo = datatable_compoundsInfo.T
        datatable_normalized.rename(columns={"Compound": "Sample"})
        datatable_raw.rename(columns={"Compound": "Sample"})
        return (
            datatable_compoundsInfo,
            datatable_normalized,
            datatable_raw,
            labels,
            sample_names,
        )
    else:
        datatable_compoundsInfo = datatable.iloc[:, 0:start_normalized]
        datatable_normalized = datatable.iloc[:, start_normalized:]
        datatable_normalized = datatable_normalized.T
        datatable_compoundsInfo = datatable_compoundsInfo.T
        datatable_normalized.rename(columns={"Compound": "Sample"})
        return datatable_compoundsInfo, datatable_normalized, labels, sample_names


def filter_sample_based_on_labels(data, labels, labels_to_keep):
    """
    function not used
    """
    labels_filter = np.array([i in labels_to_keep for i in labels])
    d = data.iloc[labels_filter]
    l = np.array(labels)[labels_filter]
    return d, l


def get_group_to_class(classes):
    """
    function not used
    """
    group_to_class = {}
    for class_name in classes:
        for subgroup in classes[class_name]:
            group_to_class[subgroup] = class_name
    return group_to_class


def reverse_dict(dictionnary: dict) -> dict:
    """
    Create a reverse dict to easily retrieve the label associate to a target.
    example
    input dict is in shape {label1 : [target1, target2], label2 : [target3, target4]}
    output dict would be {target1 : label1, target2 : label1, target3 : label2, target4 : label2}
    """
    reversed_dict = {}
    for key, value in dictionnary.items():
        if type(value) is list:
            for val in value:
                reversed_dict[val] = key
        else:
            reversed_dict[value] = key
    return reversed_dict


def load_classes_from_targets(classes_design: dict, targets: Tuple[str]) -> List[str]:
    """
    Create a list of targets to be predicted according to the class design, kind of a "traduction" of labels.
    !!! There is some confusion between class/label/target, it results from their use as synonyms in general in
    the litterature. We tried to established a distinction because we need 3 different terms to name 3 slightly different 
    things. However, even for us it got mixed up during the development. Hence the sometime confusing naming of variables.
    Always refer to the documentation for the proper meaning. !!!
    example
    Argument 'targets' is (class3, class1, class4, class4, class2, class1)
    classes_design would be something like {label1 : [class1, class2], label2 : [class3, class4]}
    reverse_classes_design would be {class1 : label1, class2 : label1, class3 : label2, class4 : label2}
    'classes' that is returned would be [label2, label1, label2, label2, label1, label1]
    """
    reverse_classes_design = reverse_dict(classes_design)
    classes = []
    for target in targets:
        if target not in reverse_classes_design:
            raise ValueError("Target {} not found in classes_design".format(target))
        classes.append(reverse_classes_design[target])
    if len(classes) != len(targets):
        raise ValueError("Some targets were not found in classes_design")
    return classes


# TODO: need to support multi-classification
def get_binary(list_to_convert: List[str], classes: List[str]) -> List[int]:
    return [classes.index(value) for value in list_to_convert]


def compute_hash(data: str) -> str:
    """
    Compute a hash for a data string
    """
    return hashlib.sha256(data.encode("utf-8")).hexdigest()


def is_save_safe(saved_metabo_experiment_dto) -> bool:
    return (
            saved_metabo_experiment_dto.metadata.is_data_the_same()
            and saved_metabo_experiment_dto.data_matrix.is_data_the_same()
    )


def format_list_for_checklist(list_to_format: List[str]) -> List[Dict[str, str]]:
    return [{"label": value, "value": value} for value in list_to_format]


def check_if_column_exist(datatable: pd.DataFrame, column_name: str) -> bool:
    return column_name in datatable.columns


def decode_pickle_from_base64(encoded_object: str):
    return pickle.loads(base64.b64decode(encoded_object.split(",")[1]))


def are_files_corresponding_to_dto(
        data: str, metadata: str, metabo_experiment_dto
) -> bool:
    return is_data_the_same(data, metabo_experiment_dto) and is_metadata_the_same(
        metadata, metabo_experiment_dto
    )


def reset_file(file_path: str):
    """
    Reset the content of a file ?
    """
    open(file_path, "w+b").close()


# TODO : function to probably delete
def restore_ids_and_targets_from_pairing_groups(filtered_samples: List[str], dataframe: pd.DataFrame, id_column: str,
                                                paired_column: str, target_column: str, classes_design: dict,) -> Tuple[List[str], List[str]]:

    pairing_values = dataframe.loc[dataframe[id_column].isin(filtered_samples)][paired_column].tolist()
    ids = dataframe[dataframe[paired_column].isin(pairing_values)][id_column].tolist()
    targets = dataframe.loc[dataframe[id_column].isin(ids)][target_column].tolist()
    duo = list(zip(ids, targets))
    restored_ids = []
    restored_targets = []
    for d in duo:
        if d[1] in np.concatenate(list(classes_design.values())):
            restored_ids.append(d[0])
            restored_targets.append(d[1])
    return restored_ids, load_classes_from_targets(classes_design, restored_targets)


def convert_str_to_list_of_lists(str_to_convert: str) -> List[List[Union[str, float, int]]]:
    first_level = []
    for find in re.findall(r'\[((([\w\'".]+,? ?)+))\] ?,?', str_to_convert):
        tmp = find[0].split(',')
        second_level = []
        for element in tmp:
            element = element.strip()
            try:
                element = int(element)
            except ValueError:
                pass
            try:
                element = float(element)
            except ValueError:
                pass
            second_level.append(element)

        first_level.append(second_level)
    return first_level


def is_data_the_same(data: str, metabo_experiment_dto) -> bool:
    return metabo_experiment_dto.data_matrix.get_hash() == compute_hash(data)


def is_metadata_the_same(metadata: str, metabo_experiment_dto) -> bool:
    return metabo_experiment_dto.metadata.get_hash() == compute_hash(metadata)


def get_model_from_import(imports_list: list, model_name: str) -> sklearn:
    """
    Import a "custom" model from sklearn
    """
    last_import = importlib.import_module("." + imports_list[0], package="sklearn")

    for next_import in imports_list[1:]:
        last_import = getattr(last_import, next_import)

    model = getattr(last_import, model_name)
    return model


def get_model_parameters(model) -> List[Tuple[str, str]]:
    parameters = vars(model()).keys()
    parameters = [(parameter, _get_type(parameter, model())) for parameter in parameters if parameter not in ["self", "random_state"]]
    return parameters


def _get_type(attribute: str, owner_instance) -> str:
    return str(type(getattr(owner_instance, attribute)).__name__)


def _parameter_is_relevant(parameter: str, model: object) -> bool:
    return parameter not in get_model_parameters(model) and parameter.endswith("_") and not parameter.startswith("_")


def _parameter_is_collection(parameter: str, trained_model: sklearn) -> bool:
    try:
        attribute = getattr(trained_model, parameter)
    except AttributeError:
        print("Attribute {} not found in model".format(parameter))
        return False
    if type(attribute) in (list, tuple, set, dict, pd.DataFrame, pd.Series, numpy.ndarray):
        if len(attribute) == 3:
            return True
        try:
            # TODO: manual test on the web app
            if len(attribute[0]) == 3 and len(attribute) == 1:
                return True
        except (KeyError, TypeError, IndexError):
            pass
    return False


def get_model_parameters_after_training(model: sklearn) -> List[Tuple[str, str]]:
    trained_model = model()
    trained_model.fit([[1, 2, 3], [4, 5, 6]], [1, 2])

    attributes = dir(trained_model)
    parameters = []
    for attribute in attributes:
        if _parameter_is_collection(attribute, trained_model) and \
                _parameter_is_relevant(attribute, model):
            parameters.append((attribute, _get_type(attribute, trained_model)))
    if DEFAULT_IMPORTANCE_ATTRIBUTE in attributes:
        default_tuple = (DEFAULT_IMPORTANCE_ATTRIBUTE, _get_type(DEFAULT_IMPORTANCE_ATTRIBUTE, trained_model))
        parameters.remove(default_tuple)
        parameters.insert(0, default_tuple)
    return parameters


def transform_params_to_cross_validation_dict(params: List[Tuple[str, str]], param_types: dict) -> dict:
    cross_validation_params = {}
    error = []
    for param, value in params:
        values = re.split(r" *, *", value)
        param_type = param_types[param]
        if param_type == "float":
            try:
                cross_validation_params[param] = [float(val) for val in values]
            except ValueError:
                error.append(f"{param} must be decimal numbers (with '.')")
        elif param_type == "int":
            try:
                cross_validation_params[param] = [int(val) for val in values]
            except ValueError:
                error.append(f"{param} must be integers")
        else:
            try:
                values = int(value)
            except ValueError:
                try:
                    values = float(value)
                except ValueError:
                    pass
            try:
                values = [int(val) for val in values]
            except ValueError:
                try:
                    values = [float(val) for val in values]
                except ValueError:
                    pass

            cross_validation_params[param] = values
    if error:
        raise ValueError(error)
    return cross_validation_params


def get_closest_integer_steps(slider_size):
    max_number_of_steps = 5

    if slider_size <= 0:
        return []
    step = slider_size // max_number_of_steps
    steps = [step * i for i in range(0, max_number_of_steps + 1)]
    if slider_size % max_number_of_steps != 0:
        # steps.pop(-1)
        steps.append(slider_size)
    return [int(i) for i in steps if i <= slider_size]


def remove_random_samples_from_class(X: pd.Series, y: List[str], balance_correction: int,
                                     classes_repartition: dict, seed: int = 42)-> Tuple[pd.Series, list]:
    """
    Function to adjust proportion between classes, to make it more balanced, or as balanced as possible
    Supposed to be provided by user (how much to adjust)
    """

    samples_ids_and_targets = pd.DataFrame({"id": X, "final_classes": y})
    balance_correction = balance_correction / 100

    if len(classes_repartition) > 2:
        raise ValueError("Balance correction is not supported for multiclassification")

    class_A_name, class_B_name = tuple(classes_repartition.keys())
    total_number_of_samples = classes_repartition[class_A_name] + classes_repartition[class_B_name]
    class_A_repartition = classes_repartition[class_A_name] / total_number_of_samples
    class_B_repartition = classes_repartition[class_B_name] / total_number_of_samples

    # Naming trick to ensure that the "A" class is always the one with the higher number of examples
    if class_B_repartition > class_A_repartition:
        class_A_name, class_B_name = class_B_name, class_A_name
        class_A_repartition, class_B_repartition = class_B_repartition, class_A_repartition

    class_A_lines = samples_ids_and_targets[samples_ids_and_targets["final_classes"] == class_A_name]

    class_A_number_of_samples = len(class_A_lines)
    class_B_number_of_samples = len(samples_ids_and_targets[samples_ids_and_targets["final_classes"] == class_B_name])

    # --- PROOF OF THE FORMULA ---
    # new_proportion_A = trimmed_classe_A_samples / (trimmed_classe_A_samples + class_B_samples)
    # proportion_A - correction = trimmed_classe_A_samples / (trimmed_classe_A_samples + class_B_samples)
    # (proportion_A - correction) * (trimmed_classe_A_samples + class_B_samples) = trimmed_classe_A_samples
    # proportion_A * trimmed_classe_A_samples - correction * trimmed_classe_A_samples + proportion_A * class_B_samples - correction * class_B_samples = trimmed_classe_A_samples
    # proportion_A * class_B_samples - correction * class_B_samples = trimmed_classe_A_samples - proportion_A * trimmed_classe_A_samples + correction * trimmed_classe_A_samples
    # proportion_A * class_B_samples - correction * class_B_samples = trimmed_classe_A_samples * (1 - proportion_A + correction)
    # trimmed_classe_A_samples = class_B_samples * (proportion_A - correction) / (1 - proportion_A + correction)

    new_class_A_number_of_samples = class_B_number_of_samples * (class_A_repartition - balance_correction) / \
                                    (1 - class_A_repartition + balance_correction)
    # Considering class A is always bigger than or equal to class B, to get a more balanced repartition,
    # "new_class_A_number_of_samples" should be a smaller number than "class_A_number_of_samples"
    number_of_samples_to_remove = class_A_number_of_samples - new_class_A_number_of_samples

    np.random.seed(seed)
    ids_to_remove = np.random.choice(class_A_lines.index, int(number_of_samples_to_remove), replace=False)
    samples_ids_and_targets = samples_ids_and_targets.drop(ids_to_remove)

    return samples_ids_and_targets["id"], samples_ids_and_targets["final_classes"].tolist()



from .InfoTab import InfoTab
from .SplitsTab import SplitsTab
from .MLTab import MLTab
from .ResultsTab import ResultsTab
from .AggregatedResultsTab import AggregatedResultsTab
from .InterpretTab import InterpretTab



from . import MetaboExperiment


class ExperimentDTO:
    def __init__(self, metabo_experiment: MetaboExperiment):
        self.data_matrix = metabo_experiment.get_data_matrix()
        self.metadata = metabo_experiment.get_metadata()

        self.number_of_splits = metabo_experiment.get_number_of_splits()
        self.train_test_proportion = metabo_experiment.get_train_test_proportion()

        self.classification_designs = metabo_experiment.get_classification_designs()

        self.custom_models = metabo_experiment.get_custom_models()
        self.selected_models = metabo_experiment.get_selected_models()

        self.selected_cv_type = metabo_experiment.get_selected_cv_type()



import importlib, os
from sklearn.model_selection import GridSearchCV
from sklearn import tree, ensemble
import pickle as pkl
from .LoggerConfig import init_logger


class runAlgo:
    def __init__(self, algo_name, cv_folds, params, algo_import=""):
        self._logger = init_logger()
        self.name = algo_name  # Name of the algorithm, ex: DecisionTreeClassifier
        self.imp = algo_import  # modules imports to get the algo
        if self.imp != "":
            self.algo = self._import_the_algorithm()
        elif self.name == "DecisionTreeClassifier":
            self.algo = getattr(tree, self.name)
        else:
            self.algo = getattr(ensemble, self.name)
        self.gs = 0
        self.params = params  # (dictionary) parameters to explore with the gridsearch
        self.cv_folds = int(cv_folds)  # number of folds to do gridsearch

    def learn(self, options_dict):
        """
        Will be the function to be parallelized through processes
        :param options_dict: Dictionary containing the split options to run an algo on a split's dataset
        :return: Nothing, saves results to file
        """
        self._logger.info("entrée dans fonction learn")
        Xtrain = options_dict[0]
        Xtest = options_dict[1]
        ytrain = options_dict[2]
        ytest = options_dict[3]
        design_name = "test_na_vs_med"  # name of the explored design, ex: ctrl_vs_sick, pos_vs_neg
        split_no = options_dict[4]  # number of the split
        # Xtrain = options_dict["Xtrain"]
        # Xtest = options_dict["Xtest"]
        # ytrain = options_dict["ytrain"]
        # ytest = options_dict["ytest"]
        # design_name = options_dict["design_name"]  # name of the explored design, ex: ctrl_vs_sick, pos_vs_neg
        # split_no = options_dict["split_no"]  # number of the split

        # do GridSearchCV and save results to file
        train_predict, test_predict = self._gridSearch(
            self.params, self.cv_folds, Xtrain, Xtest, ytrain
        )
        self._save_results_to_file(
            design_name, split_no, train_predict, test_predict, ytrain, ytest
        )

    def _import_the_algorithm(self):
        imports = self.imp.split(".")

        m = importlib.import_module("." + imports[0], package="sklearn")
        for i in imports[1:]:
            m = getattr(m, i)

        a = getattr(m, self.name)
        return a

    def _gridSearch(self, params, folds, Xtrain, Xtest, ytrain):
        self._logger.info("starting the gridsearch")
        algo = self.algo()
        self.gs = GridSearchCV(algo, params, cv=folds)
        self.gs.fit(Xtrain, ytrain)
        # Predict on train.
        train_predict = self.gs.predict(Xtrain)
        # Predict on test.
        test_predict = self.gs.predict(Xtest)
        return train_predict, test_predict

    def _save_results_to_file(
        self,
        design_name,
        split_no,
        train_predict,
        test_predict,
        train_targets,
        test_targets,
    ):
        self._logger.info("saving to file")
        # Save to file.
        with open(
            os.path.join(
                "Results", "{}_{}_{}.pkl".format(design_name, split_no, self.name)
            ),
            "wb",
        ) as fo:
            pkl.dump(self.gs, fo)
            pkl.dump(train_predict, fo)
            pkl.dump(test_predict, fo)
            pkl.dump(train_targets, fo)
            pkl.dump(test_targets, fo)



from typing import Generator, Tuple, Dict, Union

from . import SplitGroup, MetaData
from .Results import *


class ClassificationDesign:
    def __init__(self, classes_design: dict):
        self._classes_design: dict = classes_design  # dict with labels as keys and classes as values
        self._name: str = ""
        self._compute_name()
        self._split_group: Union[SplitGroup, None] = None
        self._selected_models_name: Union[list, None] = None
        self.design_Results: dict = {}
        self._is_done: bool = False
        self._balance_correction: int = 0

    def get_is_done(self) -> bool:
        """
        Return the attribute is_done which is trigered when the learning is done
        """
        return self._is_done

    def set_is_done(self, is_done: bool) -> None:
        self._is_done = is_done

    def get_balance_correction(self) -> int:
        return self._balance_correction

    def set_balance_correction(self, balance_correction: int) -> None:
        if balance_correction < 0:
            raise ValueError("Balance correction cannot be negative")
        self._balance_correction = balance_correction

    def set_split_parameter_and_compute_splits(self, train_test_proportion: float, number_of_splits: int,
                                               metadata: MetaData, pairing_column: str, uniq_sample_id: List[str],
                                               test_split_seed: int|None=None) -> None:
        """
        Retrieve the classes repartition which is needed to create an instance of SplitGroup
        Create an instance of SplitGroup in the attribute _split_group
        (The init of SplitGroup triggers the _compute_splits function).
        If test_split_seed is provided, then only this test split seed is computed.
        """
        classes_repartition = metadata.get_classes_repartition_based_on_design(self._classes_design)
        self._split_group = SplitGroup(metadata, self.get_selected_targets_name(), train_test_proportion,
                                       number_of_splits, self._classes_design, pairing_column, uniq_sample_id, 
                                       self._balance_correction, classes_repartition, test_split_seed)

    def get_name(self) -> str:
        return self._name

    def get_full_name(self) -> str:
        name = []
        for key, item_list in self._classes_design.items():
            name.append(f"{key} ({', '.join(item_list)})")
        return " versus ".join(name)

    def get_classes_design(self) -> dict:
        return self._classes_design

    def set_selected_models_name(self, selected_models_name: list) -> None:
        """
        Set the attribute self._selected_models_name
        and create the attribute self.design_Results with initialized instances of the class Results for each algorithm
        selected_models_name: list of names of models to run on data
        """
        if self._split_group is None:
            raise ValueError("Trying to set models before setting splits parameters")
        self._selected_models_name = selected_models_name
        for n in self._selected_models_name:
            self.design_Results[n] = Results(self._split_group.get_number_of_splits())

    def get_results(self) -> Dict[str, Results]:
        """
        Return the results dict (attribute) corresponding to this instance of Classification Design
        """
        if self.design_Results == {}:
            raise RuntimeError("The name of the selected models has to be set before accessing results.")
        return self.design_Results

    def _compute_name(self) -> None:
        """
        The name is made up of label1_vs_label2
        Each label refering to a group of one or more classes from the data
        """
        self._name = "_vs_".join(self._classes_design)

    def get_number_of_splits(self) -> int:
        """
        Retrieve the number of splits from SplitGroup instance
        """
        return self._split_group.get_number_of_splits()

    def all_splits(self) -> Generator[Tuple[int, list], None, None]:
        if self._split_group is None:
            raise RuntimeError(
                "Trying to access Splits before setting splits parameters"
            )
        for split_index in range(self._split_group.get_number_of_splits()):
            yield split_index, self._split_group.load_split_with_index(split_index)

    def get_selected_targets_name(self) -> list:
        """
        get the _classes_design dict in input and reverse it to have the classes as key and their corresponding labels
        as value. It is then easier to retrieve a label for a specific class
        """
        return list(Utils.reverse_dict(self._classes_design).keys())



import itertools
import os
from collections import Counter
from typing import List

import numpy as np
import pandas as pd
import sklearn
import umap
from sklearn.decomposition import PCA
from sklearn.metrics import (
    accuracy_score,
    confusion_matrix,
    precision_score,
    recall_score,
    f1_score,
    roc_auc_score,
    balanced_accuracy_score,
)

from ..service import Utils
from ..conf import parameters as cfg
from ..service import Utils, init_logger

ROOT_PATH = os.path.dirname(__file__)
DUMP_PATH = os.path.join(ROOT_PATH, os.path.join("dumps", "splits"))

MAX_CARDINALITY_FOR_COOCURENCE = 1000


class Results:
    """
    Contains all results of a classification design, is an attribute of class Classification_design, and gives info to class "Plotter".
    Has results of all algorithms for all splits on one classification design (so almost only numbers/floats/ints).
    Can be kept in RAM as it is not supposed to be too big, and prevents the reading/writing of models and splits files.
    """

    def __init__(self, splits_number: int):
        self._logger = init_logger()
        self.splits_number = [str(s) for s in range(splits_number)]
        self.results = {s: {} for s in self.splits_number}
        self.f_names = []
        self.best_acc = 0
        self.design_name = ""

        self.tmp = {"scaled_data": pd.DataFrame(), "y_train_true": [], "y_test_true": [], "classes": []}

    def add_results_from_one_algo_on_one_split(
            self,
            model: sklearn,
            scaled_data: pd.DataFrame,
            importance_attribute: str,
            classes: list,
            y_train_true: list,
            y_train_pred: list,
            y_test_true: list,
            y_test_pred: list,
            split_number: str,
            train_ids: List[str],
            test_ids: List[str],
    ):
        """
        Besoin modèle pour extraire features, features importance
        Besoin des y_true, des y_pred, des noms de samples pour le train et le test

        Fonction appelée à chq fois qu'un split a fini de rouler, pour stocker les info nécessaires à la production des
        graphique pour l'onglet résultat
        X : entièreté du dataset (autant train que test) c'est simplement pour voir le clustering de tous les individus
        """
        self.results[split_number]["y_test_true"] = y_test_true
        self.results[split_number]["y_test_pred"] = y_test_pred
        self.results[split_number]["y_train_true"] = y_train_true
        self.results[split_number]["y_train_pred"] = y_train_pred
        self.results[split_number]["train_accuracy"] = accuracy_score(
            y_train_true, y_train_pred
        )
        self.results[split_number]["test_accuracy"] = accuracy_score(
            y_test_true, y_test_pred
        )
        self.results[split_number]["balanced_train_accuracy"] = balanced_accuracy_score(
            y_train_true, y_train_pred
        )
        self.results[split_number]["balanced_test_accuracy"] = balanced_accuracy_score(
            y_test_true, y_test_pred
        )
        unique_classes = sorted(list(set(classes)))
        binary_y_train_true = Utils.get_binary(y_train_true, unique_classes)
        binary_y_train_pred = Utils.get_binary(y_train_pred, unique_classes)
        binary_y_test_true = Utils.get_binary(y_test_true, unique_classes)
        binary_y_test_pred = Utils.get_binary(y_test_pred, unique_classes)
        self.results[split_number]["train_precision"] = precision_score(
            binary_y_train_true, binary_y_train_pred
        )
        self.results[split_number]["test_precision"] = precision_score(
            binary_y_test_true, binary_y_test_pred
        )
        self.results[split_number]["train_recall"] = recall_score(
            binary_y_train_true, binary_y_train_pred
        )
        self.results[split_number]["test_recall"] = recall_score(
            binary_y_test_true, binary_y_test_pred
        )
        self.results[split_number]["train_f1"] = f1_score(
            binary_y_train_true, binary_y_train_pred
        )
        self.results[split_number]["test_f1"] = f1_score(
            binary_y_test_true, binary_y_test_pred
        )
        self.results[split_number]["train_roc_auc"] = roc_auc_score(
            binary_y_train_true, binary_y_train_pred
        )
        self.results[split_number]["test_roc_auc"] = roc_auc_score(
            binary_y_test_true, binary_y_test_pred
        )
        self.results[split_number][
            "failed_samples"
        ] = self.produce_always_wrong_samples(
            y_train_true,
            y_train_pred,
            y_test_true,
            y_test_pred,
            split_number,
            train_ids,
            test_ids,
        )
        self.results[split_number]["hyperparameters"] = model.get_params()
        if self.results[split_number]["test_accuracy"] > self.best_acc:
            self.best_acc = self.results[split_number]["test_accuracy"]
            self.results["best_model"] = model
        self.results[split_number][
            "feature_importances"
        ] = self._get_features_importance(model, importance_attribute)
        self.results[split_number]["Confusion_matrix"] = self._produce_conf_matrix(
            y_test_true, y_test_pred
        )
        self.update_tmp(scaled_data, y_train_true, y_test_true, classes)


    def compute_remaining_results_on_all_splits(self, samples_id: list):
        if len(samples_id) != len(self.tmp["classes"]):
            raise ValueError("samples_id and classes should have the same length")

        self.results["info_expe"] = self._produce_info_expe(
            self.tmp["y_train_true"], self.tmp["y_test_true"]
        )
        self.results["features_table"] = self.produce_features_importance_table()
        self.results["accuracies_table"] = self.produce_accuracy_plot_all()
        self.results["classes"] = self.tmp["classes"]
        self.results["umap_data"] = self._produce_UMAP(
            self.tmp["scaled_data"], self.results["features_table"]
        )
        self.results["3d_umap_data"] = self._produce_UMAP(
            self.tmp["scaled_data"], self.results["features_table"], n_components=3
        )
        self.results["pca_data"] = self._produce_PCA(
            self.tmp["scaled_data"], self.results["features_table"]
        )
        self.results["3d_pca_data"] = self._produce_PCA(
            self.tmp["scaled_data"], self.results["features_table"], n_components=3
        )
        self.results["metrics_table"] = self.produce_metrics_table()
        self.results[
            "features_stripchart"
        ] = self.features_strip_chart_abundance_each_class(
            self.results["features_table"], self.tmp["scaled_data"]
        )
        self.results["features_2d_and_3d"] = self.produce_features_2d_and_3d(
            self.results["features_table"], self.tmp["scaled_data"]
        )
        self.results["coocurence_matrix"] = self.produce_coocurence_matrix()
        self.results["samples_id"] = samples_id
            

    def set_feature_names(self, x: pd.DataFrame):
        """
        retrieve features name directly from datamatrix
        """
        self.f_names = list(x.columns)

    def format_name_and_associated_values(self, names, values):
        """
        Aggregate statistics for used features.
        """
        df = pd.DataFrame({"name": names, "value": values})
        # Replace 0 by Nans so we dont count them in the statistics
        df = df.replace(0, value=np.nan)
        df = df.groupby(by="name").agg(['count', 'mean', 'std']).reset_index()

        # Replace Nans back to zeros
        df = df.fillna(0)

        # This will get a dictionnary where keys are names and values is a list of [count, mean, std]
        aggregated_statistics = df.set_index("name").T.to_dict("list")

        return aggregated_statistics


    def _produce_conf_matrix(self, y_test_true: list, y_test_pred: list):
        labels = list(set(y_test_true))
        return labels, confusion_matrix(
            y_test_true, y_test_pred, labels=labels, normalize="true"
        )

    def _produce_UMAP(self, X: pd.DataFrame, features_df: pd.DataFrame, n_components: int = 2):
        features = cfg.features
        umaps = []
        for nbr in features:
            selected_feat = features_df["features"][:nbr]
            selected_x = X.loc[:, selected_feat]
            selected_x = selected_x.to_numpy()
            umap_data = umap.UMAP(n_components=n_components, init="random", random_state=13, n_jobs=1)
            umaps.append(umap_data.fit_transform(selected_x))

        # Do the umap for all used metrics
        selected_feat = features_df.loc[features_df["times_used"] > 0]["features"]
        if selected_feat.shape[0] < 3:
            selected_feat = features_df["features"][:3]
        selected_x = X.loc[:, selected_feat]
        selected_x = selected_x.to_numpy()
        umap_data = umap.UMAP(n_components=n_components, init="random", random_state=13, n_jobs=1)
        umaps.append(umap_data.fit_transform(selected_x))

        # Redo the umap but on all the data
        selected_x = X.to_numpy()
        umap_data = umap.UMAP(n_components=n_components, init="random", random_state=13, n_jobs=1)
        umaps.append(umap_data.fit_transform(selected_x))
        return umaps

    def _produce_PCA(self, X: pd.DataFrame, features_df: pd.DataFrame, n_components: int = 2):
        nbr_feat = cfg.features
        pcas = []
        labels = []

        for nbr in nbr_feat:
            selected_feat = features_df["features"][:nbr]
            x = X.loc[:, selected_feat]
            x = x.to_numpy()

            pca = PCA(n_components=n_components)
            pcas.append(pca.fit_transform(x))
            labels.append(
                {str(i): f"PC {i + 1} ({var:.1f}%)" for i, var in enumerate(pca.explained_variance_ratio_ * 100)})

        # Do the PCA for all used feature
        selected_feat = features_df.loc[features_df["times_used"] > 0]["features"]
        if selected_feat.shape[0] < 3:
            selected_feat = features_df["features"][:3]

        x = X.loc[:, selected_feat]
        pca = PCA(n_components=n_components)
        pcas.append(pca.fit_transform(x))
        labels.append({str(i): f"PC {i + 1} ({var:.1f}%)" for i, var in enumerate(pca.explained_variance_ratio_ * 100)})

        # Redo the PCA but on all the data
        x = X.to_numpy()
        pca = PCA(n_components=n_components)
        pcas.append(pca.fit_transform(x))
        labels.append({str(i): f"PC {i + 1} ({var:.1f}%)" for i, var in enumerate(pca.explained_variance_ratio_ * 100)})

        return pcas, labels

    def _produce_info_expe(self, y_train_true, y_test_true):
        """
        produce dataframe with basic information about the dataset/experiment, like number of samples and the train-test
        proprotion, the number of class, etc.
        """
        nbr_train = len(y_train_true)
        nbr_test = len(y_test_true)
        tot = nbr_train + nbr_test
        ratio_test = round(nbr_test / tot * 100)
        ratio_train = 100 - ratio_test
        nom_stats = ["Number of samples", "Number of splits", "Train-test repartition"]
        valeurs_stats = [str(tot)]
        valeurs_stats.append(len(self.splits_number))
        valeurs_stats.append(str(ratio_train) + "% - " + str(ratio_test) + "%")
        y = y_train_true + y_test_true
        c = Counter(y)
        for k in c.keys():
            nom_stats.append("Number of class {}".format(k))
            valeurs_stats.append("{}".format(c[k]))

        d = {"stats": nom_stats, "numbers": valeurs_stats}
        df = pd.DataFrame(data=d)
        return df

    def produce_features_importance_table(self):
        """
        Fonction qui réccupère les features et leurs importances de chq split sur le train et le test pour en faire un dataframe.
        Est donnée à la fonction de plotting correspondante (après que l'instance ait été complétée avec tous
        les résultats de splits)
        """
        (
            features,
            times_used_all_splits,
            importance_or_usage_or_,
            std
        ) = self._aggregate_features_info()
        # print("--> aggregating done, importances : {}".format(importance_or_usage_or_))

        d = {
            "features": features,
            "times_used": times_used_all_splits,
            "importance_usage": importance_or_usage_or_,
            "std": std
        }
        df = pd.DataFrame(data=d)
        df["times_used"] = pd.to_numeric(df["times_used"])
        df["importance_usage"] = pd.to_numeric(df["importance_usage"])
        df["std"] = pd.to_numeric(df["std"])
        df = df.sort_values(by=["importance_usage"], ascending=False)
        return df

    def produce_accuracy_plot_all(self):
        """
        Retrieve balanced accuracies from every splits on train and test to make a dataframe
        The return is used by corresponding ploting function
        """
        x_splits_num = []
        y_splits_acc = []
        traces = []
        for s in self.splits_number:
            x_splits_num.append(str(s))  # c'est normal
            x_splits_num.append(str(s))
            y_splits_acc.append(self.results[s]["balanced_train_accuracy"])
            traces.append("train")
            y_splits_acc.append(self.results[s]["balanced_test_accuracy"])
            traces.append("test")

        d = {"splits": x_splits_num, "balanced accuracies": y_splits_acc, "color": traces}
        df = pd.DataFrame(data=d)

        return df

    # TODO: faire une fonction qui produce metrics table pour tous les splits
    def produce_metrics_table(self):
        metrics = [
            "accuracy",
            "balanced accuracy",
            "precision",
            "recall",
            "f1",
            "roc_auc",
        ]
        trains_metrics = []
        tests_metrics = []
        acctrain = []
        acctest = []
        balacctrain = []
        balacctest = []
        precisiontrain = []
        precisiontest = []
        recalltrain = []
        recalltest = []
        f1train = []
        f1test = []
        roc_auc_train = []
        roc_auc_test = []
        for s in self.splits_number:
            acctrain.append(self.results[s]["train_accuracy"])
            acctest.append(self.results[s]["test_accuracy"])
            balacctrain.append(self.results[s]["balanced_train_accuracy"])
            balacctest.append(self.results[s]["balanced_test_accuracy"])
            precisiontrain.append(self.results[s]["train_precision"])
            precisiontest.append(self.results[s]["test_precision"])
            recalltrain.append(self.results[s]["train_recall"])
            recalltest.append(self.results[s]["test_recall"])
            f1train.append(self.results[s]["train_f1"])
            f1test.append(self.results[s]["test_f1"])
            roc_auc_train.append(self.results[s]["train_roc_auc"])
            roc_auc_test.append(self.results[s]["test_roc_auc"])

        trains_metrics.append(
            str(round(float(np.mean(acctrain)), 4))
            + " ("
            + str(round(float(np.std(acctrain)), 4))
            + ")"
        )
        trains_metrics.append(
            str(round(float(np.mean(balacctrain)), 4))
            + " ("
            + str(round(float(np.std(balacctrain)), 4))
            + ")"
        )
        trains_metrics.append(
            str(round(float(np.mean(precisiontrain)), 4))
            + " ("
            + str(round(float(np.std(precisiontrain)), 4))
            + ")"
        )
        trains_metrics.append(
            str(round(float(np.mean(recalltrain)), 4))
            + " ("
            + str(round(float(np.std(recalltrain)), 4))
            + ")"
        )
        trains_metrics.append(
            str(round(float(np.mean(f1train)), 4))
            + " ("
            + str(round(float(np.std(f1train)), 4))
            + ")"
        )
        trains_metrics.append(
            str(round(float(np.mean(roc_auc_train)), 4))
            + " ("
            + str(round(float(np.std(roc_auc_train)), 4))
            + ")"
        )

        tests_metrics.append(
            str(round(float(np.mean(acctest)), 4))
            + " ("
            + str(round(float(np.std(acctest)), 4))
            + ")"
        )
        tests_metrics.append(
            str(round(float(np.mean(balacctest)), 4))
            + " ("
            + str(round(float(np.std(balacctest)), 4))
            + ")"
        )
        tests_metrics.append(
            str(round(float(np.mean(precisiontest)), 4))
            + " ("
            + str(round(float(np.std(precisiontest)), 4))
            + ")"
        )
        tests_metrics.append(
            str(round(float(np.mean(recalltest)), 4))
            + " ("
            + str(round(float(np.std(recalltest)), 4))
            + ")"
        )
        tests_metrics.append(
            str(round(float(np.mean(f1test)), 4))
            + " ("
            + str(round(float(np.std(f1test)), 4))
            + ")"
        )
        tests_metrics.append(
            str(round(float(np.mean(roc_auc_test)), 4))
            + " ("
            + str(round(float(np.std(roc_auc_test)), 4))
            + ")"
        )

        metrics_table = pd.DataFrame(
            data={"metrics": metrics, "train": trains_metrics, "test": tests_metrics}
        )
        return metrics_table

    def features_strip_chart_abundance_each_class(self, feature_df, data):
        """
        store data for the 10 most important feature (mean of all split)
        as well as the class for each sample
        allows ploting the stripchart in Plots
        """
        number_of_used_features = len(feature_df[feature_df["importance_usage"] > 0])
        strip_charts = []
        n_features = cfg.features + [number_of_used_features]
        for ind in n_features:
            important_features = list(feature_df["features"])[:ind]
            df = data.loc[:, important_features]
            df["targets"] = self.results["classes"]
            strip_charts.append(df)
        return strip_charts

    def produce_always_wrong_samples(
            self,
            y_train_true,
            y_train_pred,
            y_test_true,
            y_test_pred,
            split_number,
            train_ids: List[str],
            test_ids: List[str],
    ):
        """
        return: two dicts with sample names as keys, and wrongly predicted as values (0:good pred, 1:bad pred)
        """

        train_samples = {t: 0 for t in train_ids}
        test_samples = {t: 0 for t in test_ids}

        labels = {l: idx for idx, l in enumerate(list(set(y_train_true)))}

        y_train_true = [labels[l] for l in y_train_true]
        y_train_pred = [labels[l] for l in y_train_pred]
        y_test_true = [labels[l] for l in y_test_true]
        y_test_pred = [labels[l] for l in y_test_pred]

        train_nbr = [sum(x) for x in list(zip(y_train_true, y_train_pred))]
        for i, n in enumerate(train_ids):
            if train_nbr[i] == 1:
                train_samples[n] += 1

        test_nbr = [sum(x) for x in list(zip(y_test_true, y_test_pred))]
        for i, n in enumerate(test_ids):
            if test_nbr[i] == 1:
                test_samples[n] += 1

        return train_samples, test_samples

    def produce_features_2d_and_3d(self, features_table: pd.DataFrame, scaled_data):
        selected_features = features_table[:3]["features"]
        return scaled_data.loc[:, selected_features]

    def update_tmp(self, scaled_data, y_train_true, y_test_true, classes):
        if self.tmp["scaled_data"].empty:
            self.tmp["scaled_data"] = scaled_data
        if not self.tmp["y_train_true"]:
            self.tmp["y_train_true"] = y_train_true
        if not self.tmp["y_test_true"]:
            self.tmp["y_test_true"] = y_test_true
        if not self.tmp["classes"]:
            self.tmp["classes"] = classes

    def _get_features_importance(self, model, importance_attribute):
        """
        retrieve features and their importance from a model to save it in the Results dict after each split
        """
        if self.f_names is None:
            raise RuntimeError("Features names are not retrieved yet")

        if hasattr(model, importance_attribute):
            importances = getattr(model, importance_attribute)
            if len(importances) == 1 and len(importances[0]) == len(self.f_names):
                importances = importances[0]
        elif hasattr(model, 'rule_importances_'):
            importances = [0] * len(self.f_names)
            for rule, f_importance in zip(model.model_.rules, model.rule_importances_):
                importances[rule.feature_idx] = f_importance
        zipped = zip(self.f_names, importances)
        return list(zipped)

    def _aggregate_features_info(self):
        """
        When all splits are done and saved, aggregate feature info from every split to compute stats
        from all splits, concatenate in the same list the name of features, and another list their importance
        """
        features = []
        imp = []
        # Get values of all splits in two lists
        for split in self.splits_number:
            f, i = list(zip(*self.results[split]["feature_importances"]))
            features.extend(f)
            imp.extend(i)

        # Store the mean importance, and the number of time used, per feature
        count_f = self.format_name_and_associated_values(features, imp)

        features = [f for f in count_f.keys()]
        times_used_all_splits = [count_f[f][0] for f in count_f.keys()]
        importance_or_usage_or_ = [count_f[f][1] for f in count_f.keys()]
        std = [count_f[f][2] for f in count_f.keys()]
        return features, times_used_all_splits, importance_or_usage_or_, std

    def produce_coocurence_matrix(self):

        features = list(zip(*self.results["0"]["feature_importances"]))[0]
        splits = [split for split in self.results.keys() if split.isdigit()]

        weight_matrix = pd.DataFrame(0, columns=features, index=splits)
        for split, values in self.results.items():
            if split.isdigit():
                for feature, weight in values['feature_importances']:
                    if weight > 0:
                        weight_matrix.loc[split, feature] = 1
        weight_matrix = weight_matrix.loc[:, (weight_matrix != 0).any(axis=0)]
        features = list(weight_matrix.columns)

        # Mean of importance for all features
        mean_importance = weight_matrix.mean(axis=0)
        number_of_nodes = len(mean_importance)
        cardinality = (number_of_nodes * (number_of_nodes - 1)) / 2
        if cardinality > MAX_CARDINALITY_FOR_COOCURENCE:
            return None, None, None, cardinality

        all_pairs = []
        for split in splits:
            split_features = weight_matrix.loc[split][weight_matrix.loc[split] == 1].index

            pairs = list(itertools.combinations(split_features, 2))
            all_pairs.extend(pairs)

        counter = Counter(all_pairs)
        return counter, mean_importance, len(splits), int(cardinality)

# Kept old classes for compatibility
class ResultsDT(Results):
    pass


class ResultsRF(Results):
    pass


class ResultsSCM(Results):
    pass


class ResultsRSCM(Results):
    pass



import numpy as np
from pyscm.scm import SetCoveringMachineClassifier
from randomscm.randomscm import RandomScmClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier

LEARN_CONFIG = {
    "DecisionTree": {
        "function": DecisionTreeClassifier,
        "ParamGrid": {
            "RandomSearch": {
                "max_depth": np.arange(2, 9, step=1, dtype=int),
                "min_samples_split": np.arange(2, 20, step=2, dtype=int),
                "max_features": ["sqrt", "log2"],
            },
            "GridSearch": {
                "max_depth": [2, 3, 4, 5, 6, 7],
                "min_samples_split": [2, 4, 6, 8, 10],
                "max_features": ["sqrt", "log2"],
                "class_weight": ["balanced"]
            }
        },
        "importance_attribute": "feature_importances_",
    },
    "RandomForest": {
        "function": RandomForestClassifier,
        "ParamGrid": {
            "RandomSearch" : {
                "n_estimators": np.arange(5, 300, step=10, dtype=int),
                "max_depth": np.arange(2, 6, step=1, dtype=int), #[1, 2, 3, 4, 5],
                "min_samples_split": np.arange(2, 20, step=2, dtype=int),
            },
            "GridSearch" : {
                "n_estimators": [10, 30, 50, 70, 100, 200],  # removed 5,
                "max_depth": [2, 3, 4, 5],
                "min_samples_split": [2, 4, 6, 8, 10]
            }
        },
        "importance_attribute": "feature_importances_",
    },
    "SCM": {
        "function": SetCoveringMachineClassifier,
        "ParamGrid": {
            "RandomSearch" : {
                "p": np.logspace(-2, 2, base=10, num=30),
                "max_rules": np.arange(1, 6, 1, dtype=int),
                "model_type": ["conjunction", "disjunction"],
            },
            "GridSearch" : {
                "p": np.logspace(-2, 2, base=10, num=7), #[0.01, 0.1, 1, 10],
                "max_rules": [1, 2, 3, 4, 5],
                "model_type": ["conjunction", "disjunction"],
            }
        },
        "importance_attribute": "feature_importances_",
    },
    "RandomSCM": {
        "function": RandomScmClassifier,
        "ParamGrid": {
            "RandomSearch" : {
                "p": np.logspace(-2, 2, base=10, num=30),
                "n_estimators": np.arange(5, 200, step=10, dtype=int),
                "model_type": ["conjunction", "disjunction"],
            },
            "GridSearch" : {
                "p": np.logspace(-2, 2, base=10, num=5),
                "n_estimators": [5, 10, 30, 50, 70, 100, 200],
                "model_type": ["conjunction", "disjunction"],
            },
        },
        "importance_attribute": "feature_importances_",
    },
}
